unit ODBC;

{
    ODBC Class

    This unit represents a class used to connect with ODBC data sources, run
    SQL queries on them and read and write data to them.  It internally uses
    ODBCDef.pas for the ODBC32.DLL internally defined constants, and requires
    the use of the ODBC32.DLL at run-time, for this class encapsulates the
    difficult to use ODBC API stored with in.

    There are no procedures in this class, all are functions, and for the
    functions that return boolean types, they'll return FALSE should the
    function fail, TRUE should it be successful.  Any developer using this
    class should constantly check those for success because when dealing with
    ODBC, the possibilities are endless and things can be unpredictable
    depending upon the set up of a particular system.

    Created by KRS 6/14/1999.

    In May of 2000, Randy (RB) and Dave (DM) branched this class because of a
    need to make multiple statement handles, and other, new data sources were
    starting to be utilized, and other operating systems (like Windows NT, and
    Windows 2000) so they made continual changes and fixes up until the end of
    November, 2000.

    Re-organized and I now load the Microsoft provided ODBC DLL's dynamically,
    so any program compiled with this library won't be dependant on them if you
    are not using ODBC or have situations where ODBC is not used.  Also made
    changes to allow this library to function equally well with all currently
    supported Data Sources while trying not to sacrifice one for another.
    By KRS 12/6/2000.
}

interface
{$ifdef VER140}                           // delphi 6
  {$define GTR_50}
{$endif}
{$ifdef VER150}                           // delphi 7
  {$define GTR_50}
  {$define GTR_60}
{$endif}
{$ifdef VER180}                           // delphi 2006
  {$define GTR_50}
  {$define GTR_60}
{$endif}

uses
  SysUtils, Windows, Classes, Dialogs, IniFiles,
{$ifdef GTR_50}
  DateUtils,
  StrUtils,
{$endif}
{$ifdef UseTSql}
  sql,
{$endif}
  ODBCDef,
{$ifndef NO_MESSLOG}
  MessLog,
{$endif}
{$ifdef ISASERVICE}
  SvcMgr;
{$else}
  Forms;
{$endif}

const
  // use these to try and determine if we have lost our Data connection
  // will set the fConnectionFailure flag in TODBC, which can be checked
  // using ConnectionDead();
  ConnFailMsgs: array[1..1] of string = ( 'UNABLE TO COMPLETE NETWORK REQUEST TO HOST' );
  ConnFailStates: array[1..8] of string = ( '08001',
                                            '08002',
                                            '08003',
                                            '08004',
                                            '08005',
                                            '08006',
                                            '08007',
                                            '08S01' );

type
{$ifndef GTR_50}
  TBoundArray = array of integer;
{$endif}
  // dbSkip's kinds of possible table movement in the "oType" parameter.
  TODBCOrientation = (oNext, oFirst, oLast, oPrior, oAbsol, oRelat);
  // Possible field types of the result SQL tables.
  TODBCDataTypes = (dUnknown, dChar, dInt, dFloat, dLogical, dDate, dTime, dDateTime, dBlob);
  // how catalogs are to appear in SQL statements...
  TODBCDSFileUsage = (fuFileNotSup, fuFileTable, fuFileCatalog);
  // For variables to bind string, date, time, datetime fields using BindToField.
  ODBCString = array [0..520] of byte;        // changed to byte, just in case?
  TodBytes = array of byte;

{$ifndef UseTSql}
  // Possible kinds of data sources, like Oracle, XBase and Access respectfully.
  TdbType = (dbTypeUnknown, dbTypeOracle, dbTypeMySQL, dbTypeDBF, dbTypeFireBird, dbAccess, dbExcel, dbSqlServer, dbCodeBase);
{$endif}

  // Field information stored by the ODBC Statement Class object.
  TColInfo = record
    Name: array [0..128] of char;
    Kind: TODBCDataTypes;
    ColWidth: integer;
    BufSiz: integer;
  end;
  PColInfo = ^TColInfo;

  // This class is for a list of all the columns in a result set used by the ODBC Statement Class object.
  TColInfoList = class(TList)
    procedure Clear; override;
    function AddCol(name: string; kind: TODBCDataTypes; BufLen: integer = 0): integer;
   end;

  // This is a forward declaration for the main ODBC Class.
  TODBC = class;

  // This is the ODBC Statement Class.
  TODBCStatement = class
  private
    { Private declarations }
    fParent: TODBC;
    hStmt: SQLHSTMT;

    fConnectStatus: boolean;

    ColInfo: TColInfoList;                 // Fields for the result set information.

    fBufferRec: boolean;
    FastRecBuff: array of TodBytes;        // We will dynamically allocate to the number of fields
                                           // and, for each field, allocate to the length of the field.
    fRecBuffSize: integer;
    fNoFetchScroll: boolean;

    NTSystem: boolean;
    fEmptyDataSet: boolean;

//  fRowsReturned: Cardinal;

//  fRowsAffected: integer;   did 'seem' to work?  (see DoSQL() for where I tried it)

    fSqlPosRefresh,
    fSqlPosUpdate: boolean;          // SqlSetPos allowable parameters...

    fSqlFetchAbs,
    fSqlFetchNext,
    fSqlFetchRel: boolean;          // SqlFetchScroll allowable parameters...

    fSQLAbort: boolean;
    fSQLActive: boolean;
    fLastSQL: string;
    
    procedure ExtractResultSetInfo;
    procedure ClearFastRecBuff;
    procedure ResetCursor;
    procedure CloseCursor;
{$ifndef NOTHREAD}
    function ThreadSQLExecDirect(Stmt: SQLHSTMT; SQL: string): integer;
{$endif}
    procedure TimeLog(l: cardinal; Func: string; cmd: string);
    function OpenCursor(readOnly: boolean): boolean;
    function FastOpenCursor: boolean;
    function GetODBCError: string;
    function ProcStatCmd(rc: integer): boolean;
    function GetData(col: integer; dtyp: smallint; Data: pointer; dLen: integer = 0): integer;

    constructor Create(parent: TODBC; mBufferRec, readOnly: boolean); overload;

  public
    { Public declarations }

    property Parent : TODBC read fParent;        // Added by KRS 1/17/2007.

    property SQLAbort: boolean read fSQLAbort write fSQLAbort;
    property SQLActive: boolean read fSQLActive write fSQLActive;
    property EmptyDataSet: boolean read fEmptyDataSet write fEmptyDataSet;

    property ConnectStatus: boolean read fConnectStatus;

    property ErrorText: string read GetODBCError;
    function FormatSQL(const sql: string; const Args: array of const): boolean;
    function FormatSQLse(const sql: string; const Args: array of const): boolean;
    function DoSQL(sql: string; AlwaysFetch: boolean = False; StmtPrepared: boolean = False): boolean;
    function DoSQLse(sql: string; AlwaysFetch: boolean = False; StmtPrepared: boolean = False): boolean;
    function RunSQL(sql: string; a: array of const; t: array of string): boolean;
    function RunSQLforAvailableData(catalog: string = ''; schema: string = ''; table: string = ''; tableType: string = ''): boolean;
    function SQLTabs(catalog: string = ''; schema: string = ''; table: string = ''; types: string = ''): boolean;
    function SQLStats(table: string; catalog: string = ''; schema: string = ''): boolean;
    function SQLPrimeKeys(table: string; catalog: string = ''; schema: string = ''): boolean;
    function SQLFrgnKeys(table: string; catalog: string = ''; schema: string = ''): boolean;
    function SQLGetColumns(catalog: string = ''; schema: string = ''; table: string = ''; column: string = ''): boolean;
    function GetRowID(table: string; catalog: string = ''; schema: string = ''): boolean;
    function GetTableList(var ViewList: string; Schema: string = ''; SystemTables: boolean = False): string;
    function GetFieldList(tableName: string; var Types: TBoundArray; GetType: boolean): string; overload;
    function GetFieldList(tableName: string; var gTypes: TBoundArray): string; overload;
    function GetFieldList(tableName: string): string; overload;

    function GetColDesc(field: string; var dtype, dwidth, ddec, extype: integer): boolean; overload;
    function GetColDesc(field: string; var dtype, dwidth, ddec: integer): boolean; overload;
    function GetColDescEx(field: string; var dtype, dwidth, ddec: integer): boolean;
    function GetColCount: integer;
    function GetColName(col: integer): string;
    function FindColName(field: string): integer;
    function GetDataWidth(col: integer): integer;
    function GetColWidth(col: integer): integer;
    function GetColType(col: integer): TODBCDataTypes; overload;
    function GetColType(field: string): TODBCDataTypes; overload;
    function GetDataLen(col: integer; dt: TODBCDataTypes): integer;
    function IsNull(field: string): boolean;

    function dbSkip(oType: TODBCOrientation = oNext; amount: integer = 1): boolean;

    function fetch(col: integer): string; overload;
    function fetch(field: string): string; overload;
    function bfetch(field: string; var buf: Pointer): integer;
    function fetchNu(field: string): string;
    function ifetch(field: string): integer;
    function dfetch(field: string): double;
    function cfetch(field: string): currency;
    function mfetch(field: string): integer;
    function tfetch(field: string): boolean;
    function dtfetch(field: string): TDateTime;

    function ColumnNullable(col: integer): boolean;
    function RowsMatched: integer;
    function BindToField(field: string; variable: pointer): boolean;
    function ReplaceRow: boolean;
    function BindValues(ParaNum: integer; ParaMem: pointer): boolean; overload;
    function BindValues(ParaNum: integer; ParaMem: pointer; StrLen: integer): boolean; overload;
    function Prepare(SQL: string): boolean;

    function BoundToDate(d: string): TDateTime;
    function BoundToTime(t: string): TDateTime;
    function BoundToDateTime(dt: string): TDateTime;
    function DateToBound(d: TDateTime): ODBCString;
    function TimeToBound(t: TDateTime): ODBCString;
    function DateTimeToBound(dt: TDateTime): ODBCString;

    function CheckErr: integer;
    function GetErrMsg(ndx: smallint; var eMess, eState: string): integer;
    procedure ShowErrors(rc: integer);
    function AnyODBCErrors(ndx: smallint; var errMsg: string): boolean;

    constructor Create(parent: TODBC; readOnly: boolean = True; fMode: integer = 0); overload;
    destructor Destroy; overload; override;
  end;

  // The ODBC Class object
  TODBC = class
  private
    { Private declarations }
    hEnv: SQLHENV;                     // Various ODBC32.DLL handles.
    hDBC: SQLHDBC;

{$ifndef NO_MESSLOG}
    fMess: TMessLog;
{$endif}
{$ifdef ISASERVICE}
    // set this if you execute your query in a thread
    //   AND  you are running as a Service
  {$ifndef NOTHREAD}
    fAppThread: TServiceThread;
  {$endif}
{$endif}

    fCatSep: string;
    fCatInFront: boolean;
    fDSFileUsage: TODBCDSFileUsage;    // Whether files are catalogs or tables.

    fBufferRecs: boolean;
    fNoColAttribs: boolean;

    fShowErrors: boolean;              // Whether to display ODBC errors or suppress them.

    fAllowSelectEmptySet : boolean;    // For a special mode to ensure all empty SELECT SQL runs return true (we will use "EmptyDataSet" to tell difference), added by KRS 3/1/2006.

    Stmts: array of TODBCStatement;    // A dynamic array of all possible statement handles.

    fDSConnected: boolean;
    fConnectionFailure: boolean;

    fConnString: string;
    fDatabType: TdbType;

    procedure CheckConnFail(eState, eMess: string);
    function GetDSNInfo(filePath: string): string;
    function GetODBCError: string;
    procedure AutodbType(s: string);

    constructor Create(environmentHandle: SQLHENV); overload;

  public
    { Public declarations }
    DTmr: cardinal;

    property ErrorText: string read GetODBCError;
    property CatSep: string read fCatSep;
    property CatInFront: boolean read fCatInFront;
    property DSFileUsage: TODBCDSFileUsage read fDSFileUsage;
    property ConnString: string read fConnString;
{$ifndef NO_MESSLOG}
    property Mess: TMessLog read fMess write fMess;
{$endif}
{$ifdef ISASERVICE}
  {$ifndef NOTHREAD}
    property AppThread: TServiceThread read fAppThread write fAppThread;
  {$endif}
{$endif}
    property BufferRecs: boolean read fBufferRecs write fBufferRecs;
    property NoColAttribs: boolean read fNoColAttribs write fNoColAttribs;

    property DoShowErrors: boolean read fShowErrors write fShowErrors;

    property AllowSelectEmptySet : boolean read fAllowSelectEmptySet
                                           write fAllowSelectEmptySet;

    property DSConnected: boolean read fDSConnected;

    property DatabaseType : TDBType read fDataBType;       // Added by KRS 1/17/2007.

    procedure GetDriversList(drvrNotDesc: boolean; list: TStringList);
    procedure GetDataSourcesList(datSrcNotDesc: boolean; list: TStringList; DSNtoDvr: PChar = nil);

    function FastConnect(dataSource, login, password, catalog: string;
                                    Driver: string = ''; AddSettings: string = ''): boolean;

    function ConnectionDead: boolean;
    function ConnectToDataSource(dataSource, login, password, catalog: string;
                                  commitOff: boolean = False; Driver: string = ''; AddSettings: string = ''): boolean;
    procedure DisconnectFromDataSource;
    function ReConnect(connectionString: string): boolean;

    function SetDataType(dbType: TdbType): TdbType;
    function GetDataType: TdbType;
    procedure Release;

    function CommitSQL(commit: boolean = True): boolean;

    function NewStatement(readOnly: boolean; supportOracle: boolean = True): TODBCStatement;
    procedure FreeStatement(theStmt: TODBCStatement);

    procedure ShowErrors(rc: integer);
    function AnyODBCErrors(ndx: smallint; var errMsg: string): boolean;

    constructor Create; overload;
    destructor Destroy; override;
  end;

{$ifdef LIKE_UEXPRESS}
  TCbase = class(TODBC)
  public
    constructor Create(HostName: string = '127.0.0.1'; Database: string = ''; User_name: string = ''; Password: string = '');
  end;

  TSQLStatement = class(TODBCStatement)
    public
      function skip: boolean;
    end;
{$endif}

  PInteger = ^integer;

  TMyThread = class(TThread)
  private
    hStmt: SQLHSTMT;
    hSQL: string;
    ThreadResult: Pinteger;
  protected
    procedure Execute; override; // Main thread execution (it's not actually called, the thread functions start it)...
  published
    constructor CreateThread(PriorityLevel: TThreadPriority; Done: Pinteger; stmt: SQLHSTMT; SQL: string);
    destructor Destroy; override;
  end;

function IsWindowsNT: boolean;
function CreateODBC: TODBC;

// For a global instantianted ODBC object.  One per application.
var
  ODBCMain: TODBC;

implementation

var
  // The handle to the ODBC32.DLL which is now obtained dynamically, by KRS 12/4/2000.
  hODBC32DLL: THandle;

{ ==============================================================================
     ODBC32.DLL ODBC API Functions ...

     These method pointers are initialized when the TODBC object is created, and
     they represent functions inside the "ODBC32.DLL" that will be loaded
     dynamically upon creation of this object.  Please keep these method
     pointers with the exact same case sensitive names as they are stored in the
     DLL, no exceptions.  By KRS 12/4/2000.

     I attempted the best I can to list them in the approximate order they
     should appear in a program, from start to finish.
============================================================================== }

  SQLAllocHandle: function (HandleType: smallint;
                              InputHandle: pointer;
                              var OutputHandlePtr: pointer): smallint; stdcall;
  { This first allocates the handle for ODBC environments, connections to data
    sourses or statements. }

  SQLGetEnvAttr: function (EnvironmentHandle: pointer;
                             Attribute: integer;
                             ValuePtr: pointer;
                             BufferLength: integer;
                             var StringLength: integer): smallint; stdcall;
  // Gets the current Environments Handle's properties.

  SQLSetEnvAttr: function (EnvironmentHandle: pointer;
                             Attribute: integer;
                             ValuePtr: pointer;
                             StringLength: integer): smallint; stdcall;
  // Sets the current Environments Handle's properties.

  SQLDrivers: function (EnvironmentHandle: pointer;
                          Direction: word;
                          DriverDescription: PChar;
                          BufferLength1: smallint;
                          var DescriptionLengthPtr: smallint;
                          DriverAttributes: PChar;
                          BufferLength2: smallint;
                          var AttributesLengthPtr: smallint): smallint; stdcall;
  // This gives the user access to a list of all the ODBC drivers on the system.

  SQLDataSources: function (EnvironmentHandle: pointer; Direction: word;
                              ServerName: PChar; BufferLength1: smallint;
                              var NameLength1Ptr: smallint;
                              Description: PChar; BufferLength2: smallint;
                              var NameLength2Ptr: smallint): smallint; stdcall;
  // This gives access to a list of all attachable data sources.

  SQLGetConnectAttr: function (ConnectionHandle: pointer;
                                 Attribute: integer;
                                 ValuePtr: pointer;
                                 BufferLength: integer;
                                 var StringLength: integer): smallint; stdcall;
  // Gets the current Connection Handle's properties.

  SQLSetConnectAttr: function (ConnectionHandle: pointer;
                                 Attribute: integer;
                                 ValuePtr: pointer;
                                 StringLength: integer): smallint; stdcall;
  // Sets the current Connection Handle's properties.

  SQLConnect: function (ConnectionHandle: pointer;
                          ServerName: PChar;
                          NameLength1: smallint;
                          UserName: PChar;
                          NameLength2: smallint;
                          Authentication: PChar;
                          NameLength3: smallint): smallint; stdcall;
  // Connects to a data source.

  SQLDriverConnect: function (ConnectionHandle: pointer;
                                WindowHandle: HWND;
                                InConnectionString: PChar;
                                NameLength1: smallint;
                                OutConnectionString: PChar;
                                BufferLength: smallint;
                                var StringLength: smallint;
                                DriverCompletion: smallint): smallint; stdcall;
  // Connects to a data source a different way.

  SQLBrowseConnect: function (ConnectionHandle: pointer;
                                InConnectionString: PChar;
                                NameLength1: smallint;
                                OutConnectionString: PChar;
                                BufferLength: smallint;
                                var StringLength: smallint): smallint; stdcall;
  // partial connect in steps...

  SQLGetInfo: function (ConnectionHandle: pointer;
                          InfoType: word;
                          InfoValuePtr: pointer;
                          BufferLength: smallint;
                          var StringLengthPtr: smallint): smallint; stdcall;
  { Displays information on the current driver, data source and system regarding
    what this ODBC connect can and cannot do and how it works. }

  SQLGetStmtAttr: function (StatementHandle: pointer;
                              Attribute: integer;
                              ValuePtr: pointer;
                              BufferLength: integer;
                              var StringLength: integer): smallint; stdcall;
  // Gets the current Statement Handle's properties.

  SQLSetStmtAttr: function (StatementHandle: pointer;
                              Attribute: integer;
                              ValuePtr: pointer;
                              StringLength: integer): smallint; stdcall;
  // Sets the current Statement Handle's properties.

  SQLTables: function (StatementHandle: pointer;
                         CatalogName: PChar;
                         CatalogNameLength: smallint;
                         SchemaName: PChar;
                         SchemaNameLength: smallint;
                         TableName: PChar;
                         TableNameLength: smallint;
                         TableType: PChar;
                         TableTypeLength: smallint): smallint; stdcall;
  { Returns a list of table, catalog, or schema names, and table types store in
    a specific data source (the current one). }

  SQLStatistics: function (StatementHandle: pointer;
                         CatalogName: PChar;
                         CatalogNameLength: smallint;
                         SchemaName: PChar;
                         SchemaNameLength: smallint;
                         TableName: PChar;
                         TableNameLength: smallint;
                         Unique: smallint;
                         Reserved: smallint): smallint; stdcall;

  SQLPrimaryKeys: function (StatementHandle: pointer;
                         CatalogName: PChar;
                         CatalogNameLength: smallint;
                         SchemaName: PChar;
                         SchemaNameLength: smallint;
                         TableName: PChar;
                         TableNameLength: smallint): smallint; stdcall;

  SQLForeignKeys: function (StatementHandle: pointer;
                         PKCatalogName: PChar;
                         PKCatalogNameLength: smallint;
                         PKSchemaName: PChar;
                         PKSchemaNameLength: smallint;
                         PKTableName: PChar;
                         PKTableNameLength: smallint;
                         FKCatalogName: PChar;
                         FKCatalogNameLength: smallint;
                         FKSchemaName: PChar;
                         FKSchemaNameLength: smallint;
                         FKTableName: PChar;
                         FKTableNameLength: smallint): smallint; stdcall;

  SQLExecDirect: function (StatementHandle: pointer;
                             StatementText: PChar;
                             TextLength: integer): smallint; stdcall;
  // Executes the sent SQL statement and prepares a new result set.

  SQLPrepare: function (StatementHandle: pointer;
                             StatementText: PChar;
                             TextLength: integer): smallint; stdcall;
  // Prepares the sent SQL statement for SQLExecute.

  SQLExecute: function (StatementHandle: pointer): smallint; stdcall;
  // Executes the prepared statement....

  SQLEndTran: function (HandleType: smallint;
                          Handle: pointer;
                          CompletionType: smallint): smallint; stdcall;
  // When 'auto commit' is turned off, use this to commit or rollback transactions.

  SQLNumResultCols: function (StatementHandle: pointer;
                                var ColumnCountPtr: smallint): smallint; stdcall;
  // Used to return the number of columns in the result set.

  SQLRowCount: function (StatementHandle: pointer;
                            var RowCountPtr: integer): smallint; stdcall;
  // Used to return the number of rows affected by Sql commands...


  SQLFetch: function (StatementHandle: pointer): smallint; stdcall;
  { Used to get the first record, then the next, etc. in a result set
    (sequentially). }

  SQLFetchScroll: function (StatementHandle: pointer;
                              FetchOrientation: smallint;
                              FetchOffset: integer): smallint; stdcall;
  { Provided the cursor is SCROLLABLE, can be used the get a record for a result
    set using additional record navigational options. }

  SQLGetData: function (StatementHandle: pointer;
                          ColumnNumber: word;
                          TargetType: smallint;
                          TargetValuePtr: pointer;
                          BufferLength: integer;
                          var StrLen_or_IndPtr: integer): smallint; stdcall;
  { Used to get the data from a particular column in the currently pointed to
    record. }

  SQLDescribeCol: function (StatementHandle: pointer;
                              ColumnNumber: smallint;
                              ColumnName: PChar;
                              BufferLength: smallint;
                              var NameLengthPtr: smallint;
                              var DataTypePtr: smallint;
                              var ColumnSizePtr: cardinal;
                              var DecimalDigitsPtr: smallint;
                              var NullablePtr: smallint): smallint; stdcall;
  { Used to get various information about a particular column like name and data
    type. }

  SQLColAttribute: function (StatementHandle: pointer;
                               ColumnNumber: smallint;
                               FieldIdentifier: smallint;
                               CharacterAttributePtr: PChar;
                               BufferLength: smallint;
                               var StringLengthPtr: smallint;
                               var NumericAttributePtr: longint): smallint; stdcall;
  // A later and more advanved version of "SQLDescribeCol".

  SQLColumns: function (StatementHandle: pointer;
                          CatalogName: PChar;
                          NameLength1: smallint;
                          SchemaName: PChar;
                          NameLength2: smallint;
                          TableName: PChar;
                          NameLength3: smallint;
                          ColumnName: PChar;
                          NameLength4: smallint): smallint; stdcall;
  // Used to get the column data for a given table.

  SQLSpecialColumns: function (StatementHandle: pointer;
                                 IdentifierType: word;
                                 CatalogName: PChar;
                                 NameLength1: smallint;
                                 SchemaName: PChar;
                                 NameLength2: smallint;
                                 TableName: PChar;
                                 NameLength3: smallint;
                                 Scope: smallint;
                                 Nullable: smallint): smallint; stdcall;
  { For returning a column or columns that uniquely identify each row in the
    table, such as, of example, Oracle "pseudo-columns" like ROWID that
    "SQLColumns" might not return. }

  SQLBindCol: function (StatementHandle: pointer;
                          ColumnNumber: word;
                          TargetType: smallint;
                          TargetValuePtr: pointer;
                          BufferLength: integer;
                          StrLen_or_IndPtr: pointer): smallint; stdcall;
  { Binds a particular variable to the field data in a column, so SQLFetchs
    automatically updates the value based on the current record.  Fields can be
    written to using the variable as well in conjunction with SQLSetPos. }

  SQLBindParameter: function (StatementHandle: pointer;
                              ParameterNumber: word;
                              InputOutputType: smallint;
                              ValueType: smallint;
                              ParameterType: smallint;
                              ColumnSize: integer;
                              DecimalDigits: smallint;
                              ParameterValuePtr: pointer;
                              BufferLength: integer;
                              StrLen_or_IndPtr: pointer): smallint; stdcall;
  // used like above, but binds in a parameter in an SQL statement, so
  // SQLExecute automatically uses the value of the variable...

  SQLSetPos: function (StatementHandle: pointer;
                         RowNumber: word;
                         Operation: word;
                         LockType: word): smallint; stdcall;
  { Does a lot more, but I just use it to write data in a bound column to the
    table and possibly set record locking. }

  SQLPutData: function (StatementHandle: pointer;
                          DataPtr: pointer;
                          StrLen_or_Ind: integer): smallint; stdcall;
  // Used to write data to a table, if only I can get it to work.

  SQLCloseCursor: function (StatementHandle: pointer): smallint; stdcall;
  { Used to "close a cursor" and disregard a result set.  Must be called before
    any new call to an SQL statement to free old results or operation will fail. }

  SQLDisconnect: function (ConnectionHandle: pointer): smallint; stdcall;
  // Disconnects from currently connected data source.

  SQLFreeHandle: function (HandleType: smallint;
                             Handle: SQLHANDLE): smallint; stdcall;
  // Frees any kind of handle allocated with SQLAllocHandle.

  SQLFreeStmt: function (StatementHandle: pointer;
                           Option: word): smallint; stdcall;
  // Use to free up things like bounded columns, etc..

  SQLGetDiagField: function (HandleType: smallint;
                               Handle: Pointer;
                               RecNumber: smallint;
                               DiagIdentifier: smallint;
                               DiagInfoPtr: Pointer;
                               BufferLength: smallint;
                               var StringLengthPtr: smallint): smallint; stdcall;
  // Get a diagnostic field.

  SQLGetDiagRec: function (HandleType: smallint;
                             Handle: Pointer;
                             RecNumber: smallint;
                             SqlState: PChar;
                             var NativeError: integer;
                             MessageText: PChar;
                             BufferLength: smallint;
                             var TextLength: smallint): smallint; stdcall;
  // Get a diagnostic record.

  SQLCancel: function (StatementHandle: pointer): smallint; stdcall;
  // abort the SQLExecDirect cmd...

{ =============================== ODBC API End =============================== }

function IsWindowsNT: boolean;
// This will return TRUE if running under Windows NT, otherwise FALSE.
var
  verInfo: _OSVERSIONINFOA;
begin
  verInfo.dwOSVersionInfoSize := SizeOf(verInfo);

  if(GetVersionEx(verInfo))then
    Result := (verInfo.dwPlatformId = VER_PLATFORM_WIN32_NT)
  else
    Result := False;
end;

function ToString(buf: TodBytes): string;
begin
  SetString(Result, PChar(buf), StrLen(PChar(buf)));
  Result := Trim(Result);
end;

(*
type
  TLoadInfo = record
    SQLFunc: pointer;
    SQLFName: string;
  end;

var
  ODBCLoad: array[1..1] of TLoadInfo =(
    (SQLFunc: @@SQLAllocHandle; SQLFName: 'SQLAllocHandle')
     );
*)

// this has to be outside of the TODBC class to keep support for
// the 'legacy' function CreateODBC which calls this and it is not part of any class...
function SetODBCProcs: boolean;
//var
//  TPoint: pointer;
begin
  // First, assume everything will fail.
  Result := False;

  // Dynamically load the ODBC32.DLL library, added by KRS 12/4/2000.
  hODBC32DLL := LoadLibrary('ODBC32.DLL');

  if(hODBC32DLL < HINSTANCE_ERROR)then
    Exit;
(*
  TPoint := GetProcAddress(hODBC32DLL, PChar(ODBCLoad[1].SQLFName));
  ODBCLoad[1].SQLFunc := TPoint;
  if(ODBCLoad[1].SQLFunc = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;
*)
  { Now dynamically initialize every function inside ODBC32.DLL that this unit
    uses.  Should any fail to initialize, the TODBC object will not be allowed
    to be created.  Added by KRS 12/4/2000. }
  @SQLAllocHandle := GetProcAddress(hODBC32DLL, 'SQLAllocHandle');
  if(@SQLAllocHandle = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLGetEnvAttr := GetProcAddress(hODBC32DLL, 'SQLGetEnvAttr');
  if(@SQLGetEnvAttr = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLSetEnvAttr := GetProcAddress(hODBC32DLL, 'SQLSetEnvAttr');
  if(@SQLSetEnvAttr = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLDrivers := GetProcAddress(hODBC32DLL, 'SQLDrivers');
  if(@SQLDrivers = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLDataSources := GetProcAddress(hODBC32DLL, 'SQLDataSources');
  if(@SQLDataSources = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLGetConnectAttr := GetProcAddress(hODBC32DLL, 'SQLGetConnectAttr');
  if(@SQLGetConnectAttr = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLSetConnectAttr := GetProcAddress(hODBC32DLL, 'SQLSetConnectAttr');
  if(@SQLSetConnectAttr = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLConnect := GetProcAddress(hODBC32DLL, 'SQLConnect');
  if(@SQLConnect = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLDriverConnect := GetProcAddress(hODBC32DLL, 'SQLDriverConnect');
  if(@SQLDriverConnect = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLBrowseConnect := GetProcAddress(hODBC32DLL, 'SQLBrowseConnect');
  if(@SQLBrowseConnect = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLGetInfo := GetProcAddress(hODBC32DLL, 'SQLGetInfo');
  if(@SQLGetInfo = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLGetStmtAttr := GetProcAddress(hODBC32DLL, 'SQLGetStmtAttr');
  if(@SQLGetStmtAttr = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLSetStmtAttr := GetProcAddress(hODBC32DLL, 'SQLSetStmtAttr');
  if(@SQLSetStmtAttr = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLTables := GetProcAddress(hODBC32DLL, 'SQLTables');
  if(@SQLTables = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLStatistics := GetProcAddress(hODBC32DLL, 'SQLStatistics');
  if(@SQLStatistics = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLPrimaryKeys := GetProcAddress(hODBC32DLL, 'SQLPrimaryKeys');
  if(@SQLPrimaryKeys = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLForeignKeys := GetProcAddress(hODBC32DLL, 'SQLForeignKeys');
  if(@SQLForeignKeys = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;


  @SQLExecDirect := GetProcAddress(hODBC32DLL, 'SQLExecDirect');
  if(@SQLExecDirect = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  SQLPrepare := GetProcAddress(hODBC32DLL, 'SQLPrepare');
  if(@SQLPrepare = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  SQLExecute := GetProcAddress(hODBC32DLL, 'SQLExecute');
  if(@SQLExecute = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLEndTran := GetProcAddress(hODBC32DLL, 'SQLEndTran');
  if(@SQLEndTran = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLNumResultCols := GetProcAddress(hODBC32DLL, 'SQLNumResultCols');
  if(@SQLNumResultCols = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLRowCount := GetProcAddress(hODBC32DLL, 'SQLRowCount');
  if(@SQLRowCount = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;


  @SQLFetch := GetProcAddress(hODBC32DLL, 'SQLFetch');
  if(@SQLFetch = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLFetchScroll := GetProcAddress(hODBC32DLL, 'SQLFetchScroll');
  if(@SQLFetchScroll = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLGetData := GetProcAddress(hODBC32DLL, 'SQLGetData');
  if(@SQLGetData = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLDescribeCol := GetProcAddress(hODBC32DLL, 'SQLDescribeCol');
  if(@SQLDescribeCol = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLColAttribute := GetProcAddress(hODBC32DLL, 'SQLColAttribute');
  if(@SQLColAttribute = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLColumns := GetProcAddress(hODBC32DLL, 'SQLColumns');
  if(@SQLColumns = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLSpecialColumns := GetProcAddress(hODBC32DLL, 'SQLSpecialColumns');
  if(@SQLSpecialColumns = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLBindCol := GetProcAddress(hODBC32DLL, 'SQLBindCol');
  if(@SQLBindCol = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLBindParameter := GetProcAddress(hODBC32DLL, 'SQLBindParameter');
  if(@SQLBindParameter = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLSetPos := GetProcAddress(hODBC32DLL, 'SQLSetPos');
  if(@SQLSetPos = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLPutData := GetProcAddress(hODBC32DLL, 'SQLPutData');
  if(@SQLPutData = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLCloseCursor := GetProcAddress(hODBC32DLL, 'SQLCloseCursor');
  if(@SQLCloseCursor = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLDisconnect := GetProcAddress(hODBC32DLL, 'SQLDisconnect');
  if(@SQLDisconnect = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLFreeHandle := GetProcAddress(hODBC32DLL, 'SQLFreeHandle');
  if(@SQLFreeHandle = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLFreeStmt := GetProcAddress(hODBC32DLL, 'SQLFreeStmt');
  if(@SQLFreeStmt = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLGetDiagField := GetProcAddress(hODBC32DLL, 'SQLGetDiagField');
  if(@SQLGetDiagField = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLGetDiagRec := GetProcAddress(hODBC32DLL, 'SQLGetDiagRec');
  if(@SQLGetDiagRec = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  @SQLCancel := GetProcAddress(hODBC32DLL, 'SQLCancel');
  if(@SQLCancel = nil)then
  begin
    FreeLibrary(hODBC32DLL);
    Exit;
  end;

  Result := True;
end;

// Used the create the ODBC Class object, returns 'nil' is unsuccessful.
function CreateODBC: TODBC;
var
  hEnv: SQLHENV;
  rc: integer;
begin
  Result := nil;
  if(SetODBCProcs)then
  begin
//  Didn't seem to make much difference....
//  rc := SQLSetEnvAttr(nil, SQL_ATTR_CONNECTION_POOLING, pointer(SQL_CP_ONE_PER_HENV), 0);
//  if((rc = SQL_SUCCESS) or (rc = SQL_SUCCESS_WITH_INFO))then
//  begin
    // DLL is now succefully loaded, allocate the ODBC environment handle.
    rc := SQLAllocHandle(SQL_HANDLE_ENV, SQL_NULL_HENV, hEnv);
    if((rc = SQL_SUCCESS) or (rc = SQL_SUCCESS_WITH_INFO))then
    begin
      // Ensure that ODBC engine is to the ODBC version 3.x standard at least.
      rc := SQLSetEnvAttr(hEnv, SQL_ATTR_ODBC_VERSION, pointer(SQL_OV_ODBC3), 0);
      if((rc = SQL_SUCCESS) or (rc = SQL_SUCCESS_WITH_INFO))then
        Result := TODBC.Create(hEnv)
    end;
//  end;
  end;
end;

// Disposes the memory allocation for all of the field info's. records.
procedure TColInfoList.Clear;
var
  i: integer;
begin
  for i := 0 to Count - 1 do
    FreeMem(Items[i]);
  inherited Clear;
end;

// Adds a new field (column) to the list, setting the name and type.
function TColInfoList.AddCol(name: string; kind: TODBCDataTypes; BufLen: integer = 0): integer;
var
  col: PColInfo;
begin
  GetMem(col, SizeOf(TColInfo));
  strPLCopy(col^.name, UpperCase(name), 128);
  col^.Kind := kind;
  col^.ColWidth := BufLen;
  col^.BufSiz := 0;

  Result := Add(col);
end;

constructor TMyThread.CreateThread(PriorityLevel: TThreadPriority; Done: Pinteger; stmt: SQLHSTMT; SQL: string);
begin
  inherited Create(true);                     // Create thread suspended

  Priority := TThreadPriority(PriorityLevel); // Set Priority Level
  FreeOnTerminate := False;                   // Thread DOES NOT Free itself when terminated
  // used to check the return value of SQLExecDirect()
  ThreadResult := Done;
  // these are what is passed to SQLExecDirect()
  hStmt := Stmt;
  hSQL := SQL;
  // we're ready, let it go!
  Suspended := False;                         // Continue the thread
end;

destructor TMyThread.Destroy;
begin
// This posts a message to the main form, tells us when and which thread
// is done executing.  ('Sounds' good but NOT THIS TIME, BUDDY!)
// PostMessage(frmMain.Handle, wm_ThreadDoneMsg, self.ThreadID, 0);

  inherited destroy;
end;

// Main execution for thread
procedure TMyThread.Execute;
begin
 try
// Normally you'd use the while() loop to check if thread was
// 'requested' to stop, (procedure Terminate) but in our case
// we are only doing one thing and exiting, there is no loop...

//  while(not Terminated)do
//  begin
  ReturnValue := SQLExecDirect(hStmt, PChar(hSQL), SQL_NTS);
  ThreadResult^ := ReturnValue;
//  end;

 except
  ThreadResult^ := -1;
 end;
end;

{$ifndef NOTHREAD}
// by doing this in a 'thread' it allows this part of the code to
// remain 'active' in case the sql stmt doesn't return, so we can
// SQLCancel() it by setting fSQLAbort := True ...
function TODBCStatement.ThreadSQLExecDirect(Stmt: SQLHSTMT; SQL: string): integer;
var
  myThread: TMyThread;
  rc: integer;
begin
  fSQLAbort := False;

  rc := -99;

  fSQLActive := True;
  myThread := TMyThread.CreateThread(tpHighest, @rc, hStmt, SQL);
  while(rc = -99)do
  begin
{$ifdef ISASERVICE}
    if(Assigned(fParent.fAppThread))then
      fParent.fAppThread.ProcessRequests(False);
{$else}
    Application.ProcessMessages;
{$endif}

    if(fSQLAbort)then
      SQLCancel(hStmt);

    // don't call this TOO much!
    Windows.Sleep(20);
  end;
  fSQLActive := False;
  // WaitFor thread to be done
  // (without this, due to timing, if we call this function in a loop,
  //  sometimes it is not actually finished when we re-enter it!)
  myThread.WaitFor;
  // free the thread...
  myThread.Free;

  Result := rc;
end;
{$endif}

// Legacy stuff...
// now, just use;  stmt := TODBCStatement.Create(ODBCMain, readonly);
//
// this create is only called in the NewStatement() function...
constructor TODBCStatement.Create(parent: TODBC; mBufferRec, readOnly: boolean);
begin
  inherited Create;

  // Set the ODBC parent object for this statement object.
  fParent := parent;

  fBufferRec := mBufferRec;
  fRecBuffSize := 0;

  ColInfo := TColInfoList.Create;
  fConnectStatus := OpenCursor(readOnly);

  // don't use the SQLfetchscroll() function, it don't work in ORACLE!
{$ifdef NOFETCHCHECK}
  if(fParent.fDatabType = dbTypeOracle)then
    fNoFetchScroll := fBufferRec;
{$endif}
  (*
  else
    // encountered a problem in DBFs where
    //       oNext:  rc := SQLFetchScroll(hStmt, SQL_FETCH_NEXT, 0);
    // failed when called the first time on a single result set select...
    if(fParent.fDatabType = dbTypeDBF)then
      fNoFetchScroll := True;
  *)

  // get them initialized (I know Delphi does it, but I need to be sure!)
  fSQLAbort := False;
  fSQLActive := False;
  NTSystem := False;    // IsWindowsNT;  DON'T USE THIS!   --- Why not?
end;

// this can now be called directly, no longer need to use TODBC.NewStatement...
constructor TODBCStatement.Create(parent: TODBC; readOnly: boolean = True; fMode: integer = 0);
var
  i: integer;
begin
  inherited Create;

  // Set the ODBC parent object for this statement object.
  fParent := parent;

  fBufferRec := fParent.fBufferRecs;
  fRecBuffSize := 0;

  ColInfo := TColInfoList.Create;
  if(fMode <> 0)then    // less overhead version of OpenCursor() for speed (but not always safe or compatible!)
    fConnectStatus := FastOpenCursor
  else
    fConnectStatus := OpenCursor(readOnly);

  // don't use the fetchscroll() function, it don't work in ORACLE!
{$ifdef NOFETCHCHECK}
  if(fParent.fDatabType = dbTypeOracle)then
    fNoFetchScroll := fBufferRec;
{$endif}
  (*
  else
    if(fParent.fDatabType = dbTypeDBF)then
      fNoFetchScroll := True;
  *)

  // get them initialized (I know Delphi does it, but I need to be sure!)
  fSQLAbort := False;
  fSQLActive := False;

  NTSystem := False;    // IsWindowsNT;  DON'T USE THIS!   --- Why not?

  if(fConnectStatus)then           // Can we create the statement?
  begin
    i := Low(fParent.Stmts);
    while(i <= High(fParent.Stmts))do
    begin
      if(fParent.Stmts[i] = nil)then
        break;
      Inc(i);
    end;
    if(i > High(fParent.Stmts))then                           // See if the current statement array is full.
      SetLength(fParent.Stmts, Length(fParent.Stmts) + 1);    // Increase the capacity for the statement array by one.
    fParent.Stmts[i] := self;
  end;
end;

procedure TODBCStatement.ResetCursor;
begin
  // USE THE SQLFreeStmt(hStmt, SQL_CLOSE);
  // AS IT DOESN'T RETURN ERROR IF NO OPEN CURSOR FOR STATEMENT...
  // ShowErrors(SQLCloseCursor(hStmt));    // free the cursor

  ShowErrors(SQLFreeStmt(hStmt, SQL_CLOSE)); // (same as SQLCloseCursor)
  ShowErrors(SQLFreeStmt(hStmt, SQL_UNBIND));
  ShowErrors(SQLFreeStmt(hStmt, SQL_RESET_PARAMS));
end;

procedure TODBCStatement.CloseCursor;
begin
  if(fConnectStatus)then
  begin
            // FYI: I've seen this SQLCancel take some time, even if you're not doing anything?
    ShowErrors(SQLCancel(hStmt));         // stop whatever you're doing...

    ResetCursor;

    ShowErrors(SQLFreeHandle(SQL_HANDLE_STMT, hStmt));
    hStmt := nil;
    // let everyone know we've 'left the building'!
    fConnectStatus := False;
  end;
end;

destructor TODBCStatement.Destroy;
var
  i: integer;
begin
  // find yourself in the parent class (TODBC) and set it empty...
  for i := Low(fParent.Stmts) to High(fParent.Stmts) do
    if(fParent.Stmts[i] = self)then
    begin
      fParent.Stmts[i] := nil;
      break;
    end;

  CloseCursor;

  // Free up the record buffer.
  if(fBufferRec)then
  begin
    for i := 0 to fRecBuffSize - 1 do
      FastRecBuff[i] := nil;
    FastRecBuff := nil;
  end;

  // and free the column info
  ColInfo.Clear;
  ColInfo.Free;

  inherited Destroy;
end;

function TODBCStatement.Prepare(SQL: string): boolean;
var
  rc: integer;
begin
  ResetCursor;
  //SQLFreeStmt(hStmt, SQL_UNBIND);
  //SQLCloseCursor(hStmt);           // Disregard any previous Result set.

  rc := SQLPrepare(hStmt, PChar(SQL), SQL_NTS);
  ShowErrors(rc);

  Result := ((rc = SQL_SUCCESS) or (rc = SQL_SUCCESS_WITH_INFO));
end;

function TODBCStatement.BindValues(ParaNum: integer; ParaMem: pointer): boolean;
var
  rc: integer;
begin
  rc := SQLBindParameter(hStmt, ParaNum, SQL_PARAM_INPUT, SQL_C_LONG, SQL_INTEGER, 0, 0, ParaMem, 0, nil);
  ShowErrors(rc);
  Result := ((rc = SQL_SUCCESS) or (rc = SQL_SUCCESS_WITH_INFO));
end;

function TODBCStatement.BindValues(ParaNum: integer; ParaMem: pointer; StrLen: integer): boolean;
var
  rc: integer;
begin
  rc := SQLBindParameter(hStmt, ParaNum, SQL_PARAM_INPUT, SQL_C_CHAR, SQL_CHAR, StrLen, 0, ParaMem, StrLen, nil);
  ShowErrors(rc);
  Result := ((rc = SQL_SUCCESS) or (rc = SQL_SUCCESS_WITH_INFO));
end;

// Used internally to extract information about the current Result set and place
// this information in the "ColInfo" structure so the rest of this ODBC Class
// methods can operate.
procedure TODBCStatement.ExtractResultSetInfo;
var
  name: ODBCString;
  dType: TODBCDataTypes;
  ColI: PColInfo;
  sname: string;
  mbNumData: array[0..32] of byte;
  CWidth: cardinal;
  NumData, ColWidth: longint;
  i, ci, rc: integer;
  totCols, msglen, dataType, NumDecimals, Nullable: smallint;
begin
  ColInfo.Clear;                       // Clear the table info for any previous SQL.

  // Read in the number of result columns (fields).
  SQLNumResultCols(hStmt, totCols);

  // if buffering recs, set the size of the buffer array here
  // to match the # of columns in the 'incoming' data...
  // after freeing up the memory from any previous select
  if(fBufferRec)then
  begin
    for i := 0 to fRecBuffSize - 1 do
      fastRecBuff[i] := nil;
    fastRecBuff := nil;
    SetLength(fastRecBuff, totCols);
    fRecBuffSize := totCols;
  end;

  // Read in each column (field) name and type into "ColInfo".
  for i := 1 to totCols do
  begin
    // have encountered at least one ODBC connection that didn't support SQLColAttribute, so use this flag...
    // (it's probably an issue with ODBC Ver 2.0, not supporting that function?)
    if(fParent.NoColAttribs)then
    begin
      rc := SQLDescribeCol(hStmt, i, @name, 254, msglen, dataType, CWidth, NumDecimals, Nullable);
      ShowErrors(rc);

      SetString(sname, PChar(@name), msglen);
      ColWidth := CWidth;
    end
    else
    begin
      // this works, when above sometimes fails?
      rc := SQLColAttribute(hStmt, i, SQL_DESC_NAME, @name, 254, msglen, NumData);
      ShowErrors(rc);
      SetString(sname, PChar(@name), msglen);

      NumData := 0;
      msglen := 0;
      FillChar(mbNumData, sizeof(mbNumData), #0);
      rc := SQLColAttribute(hStmt, i, SQL_DESC_DISPLAY_SIZE, @mbNumData, 32, msglen, NumData);
//      rc := SQLColAttribute(hStmt, i, SQL_DESC_DISPLAY_SIZE, pointer(mNumData), SQL_IS_INTEGER, msglen, NumData);

      ShowErrors(rc);
      ColWidth := NumData;

      NumData := 0;
      msglen := 0;
      { This below operation blows up for non-Oracle data sources, so I am
        replacing it with one that will work, and hopefully things will still
        function in Oracle, if not, let me know, by KRS 12/6/2000. }
  //    rc := SQLColAttribute(hStmt, i, SQL_DESC_TYPE, pointer(mNumData), SQL_IS_INTEGER, msglen, NumData);

      rc := SQLColAttribute(hStmt, i, SQL_COLUMN_TYPE, @mbNumData, 32, msglen, NumData);
      //rc := SQLColAttribute(hStmt, i, SQL_COLUMN_TYPE, pointer(mNumData), SQL_IS_INTEGER, msglen, NumData);
      if((rc <> SQL_SUCCESS) and (rc <> SQL_SUCCESS_WITH_INFO))then
        // If everything fails, assume things to be of type SQL_CHAR.
        dataType := SQL_CHAR
      else
        dataType := NumData;
    end;
    // fields created as varchar(max) get a width of 0 ?
    if(fParent.GetDataType = dbSqlServer)then
      if((ColWidth = 0) and (dataType = SQL_VARCHAR))then
        ColWidth := 8000;
(*
    // the DESC_DISPLAY_SIZE is all we really need !!!

    rc := SQLColAttribute(hStmt, i, SQL_DESC_LENGTH, pointer(mNumData), SQL_IS_INTEGER, msglen, NumData);
    ShowErrs(rc);
    // returns 0 when it's a numeric column...
    if(NumData = 0)then
    begin
      rc := SQLColAttribute(hStmt, i, SQL_DESC_PRECISION, pointer(mNumData), SQL_IS_INTEGER, msglen, NumData);
      ShowErrs(rc);
      if(ColWidth <> NumData)then
        ShowMessage('Size #2 Mis-Match!')
      else
        ColWidth := NumData;

      // returns # of decimals (but above include decimals, so we just
      // need to increment by one to account for the decimal point itself!)
      rc := SQLColAttribute(hStmt, i, SQL_DESC_SCALE, pointer(mNumData), SQL_IS_INTEGER, msglen, NumData);
      ShowErrs(rc);
      if(NumData <> 0)then
        ColWidth := ColWidth + 1;
    end
    else
      ColWidth := NumData;
*)
    case dataType of
      SQL_CHAR, SQL_VARCHAR,
      SQL_WCHAR, SQL_WVARCHAR,
      SQL_VARBINARY, SQL_IMAGE        : dType := dChar;

      SQL_LONGVARCHAR, SQL_WTEXT      : dType := dBlob;

      SQL_TINYINT, SQL_BIGINT,
      SQL_INTEGER, SQL_SMALLINT       : dType := dInt;

      SQL_NUMERIC, SQL_DECIMAL        : {if decDig=0 then
                                           dType := dInt
                                         else}
                                           dType := dFloat;

      SQL_FLOAT, SQL_REAL, SQL_DOUBLE : dType := dFloat;

      SQL_BIT                         : dType := dLogical;

      SQL_TYPE_DATE                   : dType := dDate;

      SQL_TIME, SQL_TYPE_TIME         : dType := dTime;

      SQL_DATETIME, SQL_TYPE_TIMESTAMP,
      SQL_TIMESTAMP                   : dType := dDateTime;

      SQL_UNKNOWN_TYPE                : dType := dUnknown;
    else
      ColWidth := 0;
      dType := dUnknown;
    end;

    ci := ColInfo.AddCol(sname, dType, ColWidth);
    ColI := ColInfo.Items[ci];

    // when buffering, bind all cols for speed...
    // set the buffer size correctly first
    if(fBufferRec)then
      // You can't bind to a 'blob', as it could be 2,000,000+ bytes or just 1 !?!?
      if(dType <> dBlob)then
      begin
        // Not sure why a column length comes back <= 0, but it does, so this helps with that...
        if(ColWidth <= 0)then
          ColWidth := 10;

        // trying to compensate for a problem when the data exceeds the field length
        // (in .dbf's that can happen!  It's when you see ***** displayed in Fox)
        //  however, it wasn't very successful!?!?!
        if((dType = dFloat) or (dType = dInt))then
          ColWidth := ColWidth + 20;

        // make it '+ 2', but pass '+ 1' to BindCol, just to be 'safe'!
        SetLength(fastRecBuff[i - 1], ColWidth + 2);

        // trying to get the bufsiz to be filled with data length, so far no luck...
        ColI^.BufSiz := ColWidth + 1;

        // '+ 1' to account for null terminator...
        rc := SQLBindCol(hStmt, word(i), SQL_CHAR, @fastRecBuff[i - 1, 0], ColWidth + 1, @ColI^.BufSiz);
        ShowErrors(rc);
      end
      else
        // ORACLE Note: this will only work if the blob field(s) is after any
        //              bound fields as you can't intermingle bound and non-bound,
        //              except if you have ALL the non-bound columns last...
        SetLength(fastRecBuff[i - 1], 1);
  end;
end;

// seems that sometimes we need to clear the
// 'bound' buffers!?!?
procedure TODBCStatement.ClearFastRecBuff;
var
  i: integer;
begin
  for i := 0 to fRecBuffSize - 1 do
    fastRecBuff[i, 0] := 0;
end;

// minimal calls needed to get the statement handle for speed...
// NOTE: WILL NOT WORK IN ALL CASES WITH ALL DATASOURCES,
//       USE WITH CAUTION!!!
function TODBCStatement.FastOpenCursor: boolean;
var
  rc: integer;
begin
  // Allocate the ODBC statement handle.
  rc := SQLAllocHandle(SQL_HANDLE_STMT, fParent.hDBC, hStmt);
  Result := ((rc = SQL_SUCCESS) or (rc = SQL_SUCCESS_WITH_INFO));
end;

// Connects ODBC Class object to a statement handle (must be done before any SQL
// statements can be run).  It is here the program tells whether the resulting
// SQL table will be Read Only or have write access. }
function TODBCStatement.OpenCursor(readOnly: boolean): boolean;
const
  ConCurTypes: array[1..3] of integer = ( SQL_CONCUR_ROWVER, SQL_CONCUR_VALUES, SQL_CONCUR_LOCK );
var
  cType, lInfo: cardinal;
  rc, rLength, i: integer;
  sRet: smallint;
  cAtt: Word;//byte;
begin
  // Allocate the ODBC statement handle.
  rc := SQLAllocHandle(SQL_HANDLE_STMT, fParent.hDBC, hStmt);
  if((rc = SQL_SUCCESS) or (rc = SQL_SUCCESS_WITH_INFO))then
  begin
    // Currently, we only set the Cursor type in the data source to be
    // scrollable (the enable full movement) and dynamic; with CONCURRENCY
    // set from Read Only if this is to have write access.  Hopefully,
    // most of the drivers this program will come across will support this.

    if(not readOnly)then
    begin
      for i := 1 to 3 do
      begin
        rc := SQLSetStmtAttr(hStmt, SQL_ATTR_CONCURRENCY, pointer(ConCurTypes[i]), 0);
        if(rc <> SQL_ERROR)then
          break;
      end;
      // if none of these work, too bad, just go on without showing any errors...
      // in most cases, it makes no difference anyway!?!
      rc := SQL_SUCCESS;
    end
    else
      rc := SQLSetStmtAttr(hStmt, SQL_ATTR_CONCURRENCY, pointer(SQL_CONCUR_READ_ONLY), 0);
    ShowErrors(rc);

    //rc := SQLSetStmtAttr(hStmt, SQL_ATTR_CONCURRENCY, pointer(SQL_CONCUR_LOCK), 0);
    //ShowErrors(rc);

    // this is supposed to be the default, but had to do it for an INFORMIX database
    // don't show errors here as you may see 'not implemented' errors, and who cares!?
    SQLSetStmtAttr(hStmt, SQL_ATTR_METADATA_ID, pointer(SQL_FALSE), SQL_IS_UINTEGER);

    (*  Doesn't appear to work with Oracle (what else is new?) ...    Used threading instead
    // Don't like the way the Select statement will lock everything when it's a 'cumbersome' table!?!
    rc := SQLSetStmtAttr(hStmt, SQL_ATTR_ASYNC_ENABLE, pointer(SQL_ASYNC_ENABLE_ON), 0);
    ShowErrors(rc);
    //rc := SQLGetStmtAttr(hStmt, SQL_ATTR_ASYNC_ENABLE, @cType, SQL_IS_INTEGER, rLength);
    //ShowErrors(rc);
    //if(cType = SQL_ASYNC_ENABLE_ON)then
    //  ShowMessage('Woo Hoo!');
    *)

    (*
    This is taken from the release notes that came with our Oracle ODBC stuff:

    When using the cursor library, calling SQLSetStmtAttr with
    attributes of SQL_ATTR_CURSOR_SENSITIVITY or SQL_ATTR_CURSOR_SCROLLABLE
    will fail with SQL_ERROR and a SQLSTATE of HY092.
    Applications should use the attributes SQL_ATTR_CURSOR_TYPE and
    SQL_ATTR_CURSOR_CONCURRENCY instead.

    rc := SQLSetStmtAttr(hStmt, SQL_ATTR_CURSOR_SCROLLABLE, pointer(SQL_NONSCROLLABLE), 0);
    ShowErrors(rc);
    *)

    ShowErrors(SQLSetStmtAttr(hStmt, SQL_ATTR_CURSOR_TYPE, pointer(SQL_CURSOR_DYNAMIC), 0));

    // For Microsoft provided drivers like "dBase Files", Access and Excel files,
    // KEYSET_DRIVEN is perfectly OK, and gives the user more power and control
    // when manuvering inside a result, but Oracle drivers only supports
    // FORWARD_ONLY.  We will default to assume Oracle support using
    // "NewStatement" in the TODBC class.  By KRS 12/6/2000.
    if((fParent.fDatabType = dbTypeOracle) or (fParent.fDatabType = dbTypeMySQL))then
    begin
      rc := SQLGetStmtAttr(hStmt, SQL_ATTR_CURSOR_TYPE, @cType, SQL_IS_INTEGER, rLength);
      ShowErrors(rc);

      // if the DYNAMIC type is NOT supported, the ODBC driver
      // sets it to KEYDRIVEN, which I don't want, so set to FORWARD...
      // STAY AWAY from KEYDRIVEN in Oracle!!!!
      // and the MyODBC seems to have some trouble when it's in STATIC, (in SQLFetchScroll ...)
      // so do the same there...
      if(((cType = SQL_CURSOR_KEYSET_DRIVEN) and (fParent.fDatabType = dbTypeOracle)) or
                       ((cType = SQL_CURSOR_STATIC) and (fParent.fDatabType = dbTypeMySQL)))then
      begin
        rc := SQLSetStmtAttr(hStmt, SQL_ATTR_CURSOR_TYPE, pointer(SQL_CURSOR_FORWARD_ONLY), 0);
        ShowErrors(rc);
      end;
    end;

    // The above set attr will not always set it to what you ask for if
    // the driver doesn't support it, so ask for the type here to
    // determine what we actually ended up getting...
    ShowErrors(SQLGetStmtAttr(hStmt, SQL_ATTR_CURSOR_TYPE, @cType, SQL_IS_INTEGER, rLength));
    case cType of
      SQL_CURSOR_STATIC:
        cAtt := SQL_STATIC_CURSOR_ATTRIBUTES1;
      SQL_CURSOR_FORWARD_ONLY:
        cAtt := SQL_FORWARD_ONLY_CURSOR_ATTRIBUTES1;
      SQL_CURSOR_KEYSET_DRIVEN:
       cAtt := SQL_KEYSET_CURSOR_ATTRIBUTES1;
      SQL_CURSOR_DYNAMIC:
        cAtt := SQL_DYNAMIC_CURSOR_ATTRIBUTES1;
      else
        cAtt := 0;    // stop whining!!
    end;
    // information on scrollable cursors for SqlSetPos()
    ShowErrors(SQLGetInfo(fParent.hDBC, cAtt, @lInfo, SizeOf(lInfo), sRet));
    if(cType = SQL_CURSOR_STATIC)then   // this overrides what the above returns...
    begin
      fSqlPosRefresh := False;
      fSqlPosUpdate := False;
    end
    else
    begin
      fSqlPosRefresh := ((lInfo and SQL_CA1_POS_REFRESH) <> 0);
      fSqlPosUpdate := ((lInfo and SQL_CA1_POS_UPDATE) <> 0);
    end;
    // ((lInfo and SQL_CA1_POS_DELETE) <> 0);
    // ((lInfo and SQL_CA1_POS_POSITION) <> 0);

    // information on cursors for SqlFetchScroll...
    // A FetchOrientation argument of SQL_FETCH_NEXT
    fSqlFetchNext := ((lInfo and SQL_CA1_NEXT) <> 0);

    // FetchOrientation arguments of
    // SQL_FETCH_FIRST, SQL_FETCH_LAST, and SQL_FETCH_ABSOLUTE
    fSqlFetchAbs := ((lInfo and SQL_CA1_ABSOLUTE) <> 0);

    // FetchOrientation arguments of
    // SQL_FETCH_PRIOR and SQL_FETCH_RELATIVE
    fSqlFetchRel := ((lInfo and SQL_CA1_RELATIVE) <> 0);
(*
    we don't use this...
    // A FetchOrientation argument of SQL_FETCH_BOOKMARK
    fSqlFetchBook := ((lInfo and SQL_CA1_BOOKMARK) <> 0);
*)

    fNoFetchScroll := (not (fSqlFetchAbs or fSqlFetchRel));

    // make sure we get ALL rows!
    // this was to attempt to fix a 'problem', but I don't think it really
    // does anything, as the default IS to read all rows...
    rc := SQLSetStmtAttr(hStmt, SQL_ATTR_MAX_ROWS, pointer(0), 0);
    ShowErrors(rc);

    (*
    This didn't seem to work!  (was trying to get a rowcount on a Select)
    rc := SQLSetStmtAttr(hStmt, SQL_ATTR_ROWS_FETCHED_PTR, @fRowsReturned, 0);
    ShowErrors(rc);
    *)

    Result := True;
  end
  else
    Result := False;
end;

// for Access, trying to find out if the Required property is set on...
function TODBCStatement.ColumnNullable(col: integer): boolean;
var
  mbNumData: array[0..32] of byte;
  NumData: longint;
  rc: integer;
  msglen: smallint;
begin
  NumData := 0;
  msglen := 0;
  FillChar(mbNumData, sizeof(mbNumData), #0);
  rc := SQLColAttribute(hStmt, col, SQL_DESC_NULLABLE, @mbNumData, 32, msglen, NumData);
  ShowErrors(rc);
  Result := (NumData <> SQL_NO_NULLS);
end;

// This doesn't do what you think, always returns 1 or 0
// I think it's something to do with rowset size, not the total
// number of rows in the table that match the query...
function TODBCStatement.RowsMatched: integer;
begin
  Result := 1;//fRowsReturned;
end;

// use DoSQL but allow for Format strings
function TODBCStatement.FormatSQL(const sql: string; const Args: array of const): boolean;
var
  sqlStr: string;
begin
  sqlStr := Format(sql, Args);
  Result := DoSQL(sqlStr);
end;

// same as DoSQL but show all errors!!
function TODBCStatement.DoSQLse(sql: string; AlwaysFetch: boolean = False; StmtPrepared: boolean = False): boolean;
{$ifndef NO_MESSLOG}
var
  ew: boolean;
{$endif}
begin
  Result := DoSQL(sql, AlwaysFetch, StmtPrepared);

  // if Result is False, and EmptyDataSet is True,
  // it's just that no matching records were found,
  // not really an error condition
  if(not Result)then
    if(not fEmptyDataSet)then
{$ifndef NO_MESSLOG}
      if(Assigned(fParent.fMess))then
      begin
        ew := fParent.DoShowErrors;
        fParent.DoShowErrors := True;

        fParent.fMess.Add(sql);
        ShowErrors(SQL_ERROR);

        fParent.DoShowErrors := ew;
      end
      else
{$endif}      
        ShowErrors(SQL_ERROR);
end;

// use DoSQL but allow for Format strings and show all errors!!
function TODBCStatement.FormatSQLse(const sql: string; const Args: array of const): boolean;
var
  sqlStr: string;
begin
  sqlStr := Format(sql, Args);
  Result := DoSQLse(sqlStr);
end;

// get the 3rd word in SQL seperated by spaces...
function Get3rd(SQL: string): string;
var
{$ifdef GTR_60}
  l, ll, i: integer;
{$else}
  l: integer;
{$endif}
begin
{$ifdef GTR_60}
  Result := '';
  l := 1;
  for i := 1 to 3 do
  begin
    ll := l;
    l := PosEx(' ', SQL, l);
    if(l = 0)then
      break
    else
      Inc(l);

    if(i = 3)then
      Result := Copy(SQL, ll, l - (ll + 1));
  end;
{$else}
  l := Pos(' ', SQL);
  l := l + Pos(' ', Copy(SQL, l + 1, Length(SQL))) + 1;
  Result := Trim(UpperCase(Copy(SQL, l, Length(SQL))));
  l := Pos(' ', Result);
  if(l > 0)then
    Result := Copy(Result, 1, l - 1);
{$endif}
end;

procedure TODBCStatement.TimeLog(l: cardinal; Func: string; cmd: string);
begin
{$ifndef NO_MESSLOG}  if(fParent.DTmr > 0)then
    if(l >= fParent.DTmr)then
      if(Assigned(fParent.fMess))then
        fParent.fMess.Flush(Format('%15s #%.2d: %.3f secs ''%s''  %s', ['*** TIMEOVR ***', 0,  l / 1000,  Func, cmd]));
{$endif}
end;

// just run the statement, no % substitution...
function TODBCStatement.DoSQL(sql: string; AlwaysFetch: boolean = False; StmtPrepared: boolean = False): boolean;
var
{$ifdef THISCRAP}
  sTemp: WideString;
{$endif}
  Cmd: string;
  tmr: cardinal;
  rc: integer;
begin
  if(not StmtPrepared)then
    SQLFreeStmt(hStmt, SQL_UNBIND);

  // USE THE SQLFreeStmt(hStmt, SQL_CLOSE);
  // AS IT DOESN'T RETURN ERROR IF NO OPEN CURSOR FOR STATEMENT...
  // ShowErrors(SQLCloseCursor(hStmt));    // free the cursor
  ShowErrors(SQLFreeStmt(hStmt, SQL_CLOSE)); // (same as SQLCloseCursor)

  fLastSQL := SQL;
  tmr := GetTickCount();

{$ifdef THISCRAP}
  if(NTSystem)then
  begin
    // This is NOT being used at this time...
    sTemp := WideString(SQL);
    rc := SQLExecDirect(hStmt, PChar(sTemp), SQL_NTS);
  end
  else
  begin
{$endif}
    // this is for binding parameter stuff (where you put ? into the select and then
    // change what is substuted for the ?)
    if(StmtPrepared)then
      rc := SQLExecute(hStmt)
    else
    // Run the SQL statement in the ODBC engine.  (as a thread or not...)
{$ifdef NOTHREAD}
    rc := SQLExecDirect(hStmt, PChar(SQL), SQL_NTS);
{$else}
    rc := ThreadSQLExecDirect(hStmt, SQL);
{$endif}

{$ifdef THISCRAP}
  end;
{$endif}
  TimeLog(GetTickCount() - tmr, 'DoSQL', SQL);

  // sometimes we may need to know the sql statement was fine, just no data found...
  fEmptyDataSet := False;
  Cmd := Trim(UpperCase(Copy(SQL, 1, 6)));
  Result := False;
  if((rc = SQL_SUCCESS) or (rc = SQL_SUCCESS_WITH_INFO))then
  begin                                                 // if successful ...
//  if(Cmd = 'UPDATE')then
//    SQLRowCount(hStmt, fRowsAffected);

    ExtractResultSetInfo;
    // on select, may as well go top.  rb (who?) May 11/00
    // for remotedb.dll, stored procedures won't fetch
    if((Cmd = 'SELECT') or AlwaysFetch)then
    begin
      if(UpperCase(Get3rd(SQL)) = 'INTO')then
        // for statements like "select * into xx from yy"
        // where we are creating a data file not selecting anything...
        // NOTE: if the select is; 'select fred, barney into xx from yy', it
        //       WILL NOT WORK, due to the space between fred and barney...
        //       The select will succeed and the table will be created, but
        //       the dbskip below will cause an error condition...
        Result := True
      else
      begin
        Result := dbSkip(oNext, 1);
        fEmptyDataSet := (Result = False);

        // override any 'Error in row' error condition
        // if(not Result)then
        // begin
        //   if(CheckErr = SQL_NEED_DATA)then
        //     Result := True;
        // end;

        { A special mode to ensure all empty SELECT SQL runs return true (we
          will use "EmptyDataSet" to tell difference), added by KRS 3/1/2006. }
        if (not Result) then
          if (fParent.fAllowSelectEmptySet) then
            Result := True;
      end;
    end
    else
      Result := True;

//  if(Cmd = 'UPDATE')then
//  begin
//    SQLFreeStmt(hStmt, SQL_UNBIND);
//    SQLFreeStmt(hStmt, SQL_CLOSE);
//  end;
  end
  else
    if(rc = SQL_NO_DATA)then  // not an 'error', just that there were no rows to Delete or Update (Insert added here too by KRS 3/1/2006) ...
    begin
      fEmptyDataSet := True;
      if((Cmd = 'INSERT') or (Cmd = 'DELETE') or (Cmd = 'UPDATE'))then
        Result := True;
    end;
end;


function TODBCStatement.RunSQL(sql: string; a: array of const; t: array of string): boolean;
{ This function take the ODBC object to the last level by running the SQL
  statement in "SQL" and connecting to its result table.  The field names and
  types in the resulting table are stored internally in the object.  "SQL" is
  actually a string with formatting character as specified below (similar to
  Delphi's FORMAT statement) for generating generic, compatible SQL that can be
  read by most ODBC Data Sources.  To variables for these formatting specifiers
  (flags) are sent to the function in the first parameter array "A", and "T" is
  a special parameter array for sending table information (that cannot be sent
  in the first array).  Formatting flags are ...
         %%   -    %              - inserts a single "%" into the SQL string.
         %*s  -    Literal String - a SINGLE QUOTE string used as a value only.
         %s   -    String         - a DOUBLE QUOTE string used for field names, etc..
         %i   -    Integer        - inserts a plain integer into the SQL string.
         %f   -    Float          - same as above, but for a floating point value.
         %l   -    Logical        - boolean value, will insert TRUE or FALSE.
         %d   -    Date           - will insert just a date value.
         %v   -    String         - an UNQUOTED string used for whatever you want
         %#t  -    Table          - where # is the number of the table to be refered to
                                    in this SQL statement, (ie. tabel 1 is the first table
                                    listed in the "T" parameter array, table 2 is the
                                    second, etc.).
  All values in the "A" parameter array remain in their native formats; while
  the values in the "T" parameter array of tables always come in the form of a
  string formatted as "Catalog Name,Schema Name,Table Name" where no additional
  spaces are allowed in the string (except is they appear in the middle of the
  actual name and that all the commas must remain in the string, even if no
  catalog or schema is used (table is manditory).  That means if we just need
  the single table name called [Table Name] to be refered to as, use
  ",,Table Name" with no additional spaces. }
var
  newSQL, flag, cat, sch, tab: string;
  i, iA, iT, i2: integer;
begin
  { This section will turn the "SQL" with format specifiers into a machine
    readable SQL stored in the variable "newSQL", as explained above. }
  newSQL:='';
  iA:=0;
  i:=1;
  while i<=Length(SQL) do          // Navigate each character in "SQL".
    begin
      if SQL[i]<>'%' then
        // A regular character, just insert it into the new SQL statement.
        newSQL:=newSQL+SQL[i]
      else
        { An '%' was encounter, must be a format specifier (or the statement
          will be declared invalid. }
        begin

          // Read in the format specifier flag.
          flag:='';
          while (i<=Length(SQL)) and (((SQL[i]>#47) and (SQL[i]<#58)) or
           ((SQL[i]='t') and ((SQL[i-1]>#47) and (SQL[i-1]<#58))) or
           ((SQL[i]='s') and (flag='%*')) or (Length(flag)<2)) do
            begin
              flag:=flag+SQL[i];
              i:=i+1;
            end;

          // If the "%%" flag, insert a "%" and continue.
          if flag='%%' then
            begin
              newSQL:=newSQL+'%';
              continue;
            end;

          // A different kind of flag.
          i:=i-1;
          if flag[Length(flag)]='t' then
            begin
              { A Table flag, ensure it is valid and extract the catalog,
                schema and table name from the "T" parameter name, and
                insert a properly formatted table reference into the SQL. }
              iT:=StrToIntDef(Copy(flag, 2, Length(flag)-2), 0);
              if (iT=0) or (iT>High(T)+1) then         // Ensure the table number is in the array.
                begin
                  Result:=False;
                  exit;
                end;
              flag:=T[iT-1];
              i2:=Pos(',', flag);  // Get catalog name.
              if i2=0 then
                begin
                  Result:=False;
                  exit;
                end;
              cat:=Copy(flag, 1, i2-1);
              Delete(flag, 1, i2);
              i2:=Pos(',', flag);  // Get schema name.
              if i2=0 then
                begin
                  Result:=False;
                  exit;
                end;
              sch:=Copy(flag, 1, i2-1);
              Delete(flag, 1, i2);
              tab:=flag;           // Get table name.
              if (Pos(',', tab)<>0) or (tab='') then
                begin
                  Result:=False;
                  exit;
                end;

              if (cat<>'') and (fParent.CatInFront) then
                newSQL:=newSQL+'"'+cat+'"'+fParent.CatSep;
              if sch='' then
                newSQL:=newSQL+'"'+tab+'"'
              else
                newSQL:=newSQL+'"'+sch+'"."'+tab+'"';
              if (cat<>'') and (not fParent.CatInFront) then
                newSQL:=newSQL+fParent.CatSep+'"'+cat+'"';
            end

          else
            begin
              // A non-table flag that uses values from the "A" param. array.
              if iA>High(A) then                       // Ensure a corresponding value is in the array.
                begin
                  Result:=False;
                  exit;
                end
              // Literal String
              else if flag='%*s' then
                begin
                  if A[iA].VType=vtChar then
                    newSQL:=newSQL+''''+A[iA].VChar+''''
                  else if A[iA].VType=vtPChar then
                    newSQL:=newSQL+''''+A[iA].VPChar+''''
                  else
                    newSQL:=newSQL+''''+string(A[iA].VString)+'''';
                end
              // String
              else if flag='%s' then
              begin
                if A[iA].VType=vtChar then
                  newSQL:=newSQL+'"'+A[iA].VChar+'"'
                else if A[iA].VType=vtPChar then
                  newSQL:=newSQL+'"'+A[iA].VPChar+'"'
                else
                  newSQL:=newSQL+'"'+string(A[iA].VString)+'"';
              end
              // UN-Quoted string
              else if(flag='%v')then
              begin
                if(A[iA].VType = vtChar)then
                  newSQL := newSQL + A[iA].VChar
                else
                  if(A[iA].VType = vtPChar)then
                    newSQL := newSQL + A[iA].VPChar
                  else
                    newSQL := newSQL + string(A[iA].VString);
              end
              // Integer
              else if flag='%i' then
                newSQL:=newSQL+IntToStr(A[iA].VInteger)
              // Floating point value
              else if flag='%f' then
                begin
                  if A[iA].VType=vtInteger then
                    newSQL:=newSQL+FloatToStr(A[iA].VInteger)
                  else
                    newSQL:=newSQL+FloatToStr(A[iA].VExtended^);
                end
              // Logical (boolean)
              else if flag='%l' then
                begin
                  if A[iA].VBoolean then
                    newSQL:=newSQL+'TRUE'
                  else
                    newSQL:=newSQL+'FALSE';
                end
              // Date
              else if flag='%d' then
                begin
                  if A[iA].VType=vtInteger then
                    newSQL:=newSQL+'{d '''+FormatDateTime('yyyy-mm-dd', A[iA].VInteger)+'''}'
                  else
                    newSQL:=newSQL+'{d '''+FormatDateTime('yyyy-mm-dd', A[iA].VExtended^)+'''}';
                end
              // Anything else and the flag is invalid, declare the whole statement invalid.
              else
                begin
                  Result:=False;
                  exit;
                end;
              iA:=iA+1;
            end;
        end;
      i:=i+1;                      // Increment another character in "SQL".
    end;

  Result := DoSQL(newSQL);
end;

// we don't actually need to change para, but it is a
// var so that we are passing an address to the original string,
// otherwise the PChar we return would be meaningless!?
function ifNil(var para: string): pointer;
begin
  if(Length(para) > 0)then
    Result := PChar(para)
  else
    Result := nil;
end;

function ifNilT(para: string): integer;
begin
  if(Length(para) > 0)then
    Result := SQL_NTS
  else
    Result := 0;
end;

// standard routine called when processing statistical stuff,
// gets the result set ready and fetches the first row...
function TODBCStatement.ProcStatCmd(rc: integer): boolean;
begin
  if((rc = SQL_SUCCESS) or (rc = SQL_SUCCESS_WITH_INFO))then
  begin
    ExtractResultSetInfo;
    Result := dbSkip(oNext, 1);
//  rc := SQLFetch(hStmt);
//  Result := ((rc = SQL_SUCCESS_WITH_INFO) or (rc = SQL_SUCCESS));
  end
  else
    Result := False;
end;

{ This will generate a special specific result set with five columns to display
  information on available tables and schemas (as well as type info. and
  possibly catalogs) for the current Data Source.  A programmer would use this
  to show a user which available tables exist in a specific catalog, and
  functions like that.  The Result set works like in RunSQL result set except
  its fields names are ...
                   1    -    TABLE_CAT      VarChar
                   2    -    TABLE_SCHEM    VarChar
                   3    -    TABLE_NAME     VarChar
                   4    -    TABLE_TYPE     VarChar
                   5    -    REMARKS        VarChar
  I put this function in as a more raw format rather than an easy table
  returning function, because I can see the need to use this for a lot more
  than Data Sources like Access, where schema and catalog information will be
  needed, and the Result set style of doing thing gave the most information. }
function TODBCStatement.RunSQLforAvailableData(catalog, schema, table, tableType: string): boolean;
begin
  ResetCursor;          // Disregard any previous Result set.
  Result := ProcStatCmd(SQLTables(hStmt, ifNil(catalog), ifNilT(catalog), ifNil(schema), ifNilT(schema), ifNil(table), ifNilT(table), ifNil(tableType), ifNilT(tableType)));
end;

// as above, just a shorter name...
function TODBCStatement.SQLTabs(catalog, schema, table, types: string): boolean;
begin
  Result := RunSQLForAvailableData(catalog, schema, table, types);
end;

// This will generate a special specific result set with five columns to display
// information on a table's indexes etc
function TODBCStatement.SQLStats(table, catalog, schema: string): boolean;
begin
  ResetCursor;          // Disregard any previous Result set.
  Result := ProcStatCmd(SQLStatistics(hStmt, ifNil(catalog), ifNilT(catalog), ifNil(schema), ifNilT(schema), ifNil(table), ifNilT(table), SQL_INDEX_ALL, SQL_QUICK));
end;

// This will generate a special specific result set with five columns to display
// information on a table's primary keys
function TODBCStatement.SQLPrimeKeys(table, catalog, schema: string): boolean;
begin
  ResetCursor;          // Disregard any previous Result set.
  Result := ProcStatCmd(SQLPrimaryKeys(hStmt, ifNil(catalog), ifNilT(catalog), ifNil(schema), ifNilT(schema), ifNil(table), ifNilT(table)));
end;

// This will generate a special specific result set with five columns to display
// information on a table's foreign keys
function TODBCStatement.SQLFrgnKeys(table, catalog, schema: string): boolean;
begin
  ResetCursor;          // Disregard any previous Result set.
  Result := ProcStatCmd(SQLForeignKeys(hStmt, nil, 0, nil, 0, nil, 0,
                  ifNil(catalog), ifNilT(catalog), ifNil(schema), ifNilT(schema), ifNil(table), ifNilT(table)));
end;

function TODBCStatement.SQLGetColumns(catalog, schema, table, column: string): boolean;
begin
  ResetCursor;          // Disregard any previous Result set.
  Result := ProcStatCmd(SQLColumns(hStmt, ifNil(catalog), ifNilT(catalog), ifNil(schema), ifNilT(schema), ifNil(table), ifNilT(table), ifNil(column), ifNilT(column)));
end;

//      is ROWID valid to ask for with this table/view?
// actually that's not really what this does?
// what we are asking for is what would be the best way to identify a row, which usually
// with Oracle, would be a rowid, but in other cases, it may return a key field, or a
// combination of several fields that will uniquely identify a row...
// you have to skip through the result set and fiqure it out yourself, buddy!
function TODBCStatement.GetRowID(table, catalog, schema: string): boolean;
begin
  ResetCursor;          // Disregard any previous Result set.
  Result := ProcStatCmd(SQLSpecialColumns(hStmt, SQL_BEST_ROWID, ifNil(catalog), ifNilT(catalog), ifNil(schema), ifNilT(schema), ifNil(table), ifNilT(table), SQL_NO_NULLS, SQL_SCOPE_TRANSACTION));
  //Result := ProcStatCmd(SQLSpecialColumns(hStmt, SQL_BEST_ROWID, nil, 0, nil, 0, PChar(Table), SQL_NTS, SQL_NO_NULLS, SQL_SCOPE_TRANSACTION));
end;

procedure AddinString(var str: string; value: string);
begin
  if(Length(str) > 0)then
    str := str + ',';
  if(Pos(' ', value) > 0)then
    value := '"' + value + '"';
  str := str + value;
end;

const
  STR_LEN  = 127;

// returns all the tables in a database (as a comma separated string)
// and any 'views' will be in ViewList (again, comma separated)
function TODBCStatement.GetTableList(var ViewList: string; Schema: string = ''; SystemTables: boolean = False): string;
var
  Field2, Field3, Field4: array [0..STR_LEN] of char;
  Size2, Size3, Size4: integer;
  Value, TT, sch: string;
  rc: integer;
  Table, View: boolean;
begin
  ResetCursor;          // Disregard any previous Result set.
  rc := SQLTables(hStmt,
                    nil, 0,
                    ifNil(Schema), ifNilT(Schema),
                    nil, 0,
                    nil, 0);

  Result := '';
  if((rc = SQL_SUCCESS) or (rc = SQL_SUCCESS_WITH_INFO))then
  begin
    // bind the data to TableName, and type skip through the result set and you're done...
    // 2nd is schema
    // the 3rd column will have the names in it...
    // and column 4 is table type (like System or view or ...)
    ShowErrors(SQLBindCol(hStmt, word(2), SQL_CHAR, @Field2, STR_LEN - 1, @Size2));
    ShowErrors(SQLBindCol(hStmt, word(3), SQL_CHAR, @Field3, STR_LEN - 1, @Size3));
    ShowErrors(SQLBindCol(hStmt, word(4), SQL_CHAR, @Field4, STR_LEN - 1, @size4));
    rc := SQLFetch(hstmt);

    while(rc = SQL_SUCCESS)do
    begin
      Table := False;
      View := False;
      Value := string(Field3);
      TT := LowerCase(string(Field4));
      if(SystemTables)then
      begin
        sch := string(Field2);
        if(Length(sch) > 0)then
          Value := sch + '.' + Value;

        if(Pos('table', TT) > 0)then
          Table := True
        else
          if(Pos('view', TT) > 0)then
            View := True;
      end
      else
        if(TT = 'table')then
          Table := True
        else
          if(TT = 'view')then
            View := True;

      if(Table)then
        AddInString(Result, Value)
      else
        if(View)then
          AddInString(ViewList, Value);

      rc := SQLFetch(hstmt);
    end;
  end;

  // un bind it!
  SQLFreeStmt(hStmt, SQL_UNBIND);
end;

function TODBCStatement.GetFieldList(tableName: string): string;
var
  gTypes: TBoundArray;
begin
  Result := GetFieldList(tableName, gTypes, False);
end;

function TODBCStatement.GetFieldList(tableName: string; var gTypes: TBoundArray): string;
begin
  Result := GetFieldList(tableName, gTypes, True);
end;

// returns all the fields in a table (as a comma separated string)
//
// !*!*!*!*!  WARNING !*!*!*!*!
// I have discovered that some datasources (ie FireBird)
// are CASE SENSATIVE and will not find lowercase names
//
function TODBCStatement.GetFieldList(tableName: string; var Types: TBoundArray; GetType: boolean): string;
var
  rc, ColNamSiz, ColumnType: integer;
  ColumnName: array [0..STR_LEN] of char;
begin
  ResetCursor;          // Disregard any previous Result set.

  if(fParent.GetDataType = dbTypeFireBird)then
    tableName := UpperCase(tableName);

  rc := SQLColumns(hStmt,
                   nil, 0,                      // All catalogs
                   nil, 0,                      // All schemas
                   PChar(TableName), SQL_NTS,   // table name
                   nil, 0);                     // All columns

  Result := '';
  if((rc = SQL_SUCCESS) or (rc = SQL_SUCCESS_WITH_INFO))then
  begin
    // bind the data to ColumnName, skip through the result set and you're done...

    // the 4th column will have the names in it...
    ShowErrors(SQLBindCol(hStmt, word(4), SQL_CHAR, @ColumnName, STR_LEN - 1, @ColNamSiz));
    if(GetType)then      //  column 5 is data type (like SQL_DESC_TYPE)
      ShowErrors(SQLBindCol(hStmt, word(5), SQL_INTEGER, @ColumnType, sizeof(ColumnType), @ColNamSiz));

    rc := SQLFetch(hstmt);
    while(rc = SQL_SUCCESS)do
    begin
      AddInString(Result, UpperCase(string(ColumnName)));
      if(GetType)then
      begin
        SetLength(Types, Length(Types) + 1);
        Types[High(Types)] := ColumnType;
      end;
      rc := SQLFetch(hstmt);
    end;
  end;

  // un bind it!
  SQLFreeStmt(hStmt, SQL_UNBIND);
end;

// Returns the number of columns (fields) in the result set.
function TODBCStatement.GetColCount: integer;
begin
  Result := ColInfo.Count;
end;

// Returns a column name for the specified col number (starts at zero for first).
function TODBCStatement.GetColName(col: integer): string;
begin
  if(col < ColInfo.Count)then
    Result := TColInfo(ColInfo.Items[col]^).Name
  else
    Result := '';
end;

// Returns the column number of a given field name...
function TODBCStatement.FindColName(field: string): integer;
var
  i: integer;
begin
  Result := -1;
  field := UpperCase(field);
  for i := 0 to ColInfo.Count - 1 do
    if(field = PColInfo(ColInfo.Items[i])^.Name)then
    begin
      Result := i;
      break;
    end;
end;

// Returns a column type for the specified col number (starts at zero for first).
function TODBCStatement.GetColType(col: integer): TODBCDataTypes;
begin
  if((col >= 0) and (col < ColInfo.Count))then
    Result := TColInfo(ColInfo.Items[col]^).Kind
  else
    Result := dUnknown;
end;

// Returns a column type for the specified field name.
function TODBCStatement.GetColType(field: string): TODBCDataTypes;
begin
  Result := GetColType(FindColName(field));
end;

function TODBCStatement.GetColWidth(col: integer): integer;
begin
  if((col >= 0) and (col < ColInfo.Count))then
    Result := TColInfo(ColInfo.Items[col]^).ColWidth
  else
    Result := -1;
end;

function TODBCStatement.GetDataWidth(col: integer): integer;
var
  ColI: PColInfo;
begin
  if((col >= 0) and (col < ColInfo.Count))then
  begin
    ColI := ColInfo.Items[col];
    Result := ColI^.BufSiz;
  end
  else
    Result := 0;
end;

// Returns column description for the specified field name,
// formatted such that it could be used by codebase to create a dbf...
function TODBCStatement.GetColDescEx(field: string; var dtype, dwidth, ddec: integer): boolean;
var
  name: ODBCString;
  mbNumData: array[0..32] of byte;
  CWidth: cardinal;
  NumData: longint;
  i, rc: integer;
  msglen, dataType, NumDecimals, Nullable: smallint;
begin
  Result := false;
  i := FindColName(field);
  if(i >= 0)then
  begin
    i := i + 1;
    NumData := 0;
    msglen := 0;
    if(fParent.NoColAttribs)then
    begin
      rc := SQLDescribeCol(hStmt, i, @name, 254, msglen, dataType, CWidth, NumDecimals, Nullable);
      ShowErrors(rc);
      dwidth := CWidth;
      ddec := NumDecimals;
    end
    else
    begin
      FillChar(mbNumData, sizeof(mbNumData), #0);
      rc := SQLColAttribute(hStmt, i, SQL_COLUMN_TYPE, @mbNumData, 32, msglen, NumData);
//      rc := SQLColAttribute(hStmt, i, SQL_COLUMN_TYPE, pointer(mNumData), SQL_IS_INTEGER, msglen, NumData);
      if((rc <> SQL_SUCCESS) and (rc <> SQL_SUCCESS_WITH_INFO))then
        // If everything fails, assume things to be of type SQL_CHAR.
        dataType := SQL_CHAR
      else
        dataType := NumData;
    end;

    case dataType of
      SQL_CHAR,
      SQL_VARCHAR,
      SQL_VARBINARY,
      SQL_LONGVARCHAR:
        dtype := integer('C');

      SQL_TIME,
      SQL_TYPE_TIME:
        dtype := integer('H');

      SQL_DATETIME,
      SQL_TIMESTAMP,
      SQL_TYPE_TIMESTAMP:
        dtype := integer('T');

      SQL_TYPE_DATE:
        dtype := integer('D');

      SQL_BIT:
        dtype := integer('L');

      SQL_INTEGER,
      SQL_TINYINT,
      SQL_SMALLINT,
      SQL_BIGINT,
      SQL_NUMERIC:
        dtype := integer('N');

      SQL_DECIMAL,
      SQL_FLOAT,
      SQL_REAL,
      SQL_DOUBLE:
        dtype := integer('F');

      else
        // SQL_UNKNOWN_TYPE and others, default to char...
        dtype := integer('C');
    end;

    if(not fParent.NoColAttribs)then
    begin
      ddec := 0;
      if(dtype <> integer('C'))then
      begin
        rc := SQLColAttribute(hStmt, i, SQL_COLUMN_PRECISION, @mbNumData, 32, msglen, NumData);
        //rc := SQLColAttribute(hStmt, i, SQL_COLUMN_PRECISION, pointer(mNumData), SQL_IS_INTEGER, msglen, NumData);
        ShowErrors(rc);
        dwidth := NumData;

        // returns # of decimals
        rc := SQLColAttribute(hStmt, i, SQL_COLUMN_SCALE, @mbNumData, 32, msglen, NumData);
        //rc := SQLColAttribute(hStmt, i, SQL_COLUMN_SCALE, pointer(mNumData), SQL_IS_INTEGER, msglen, NumData);
        ShowErrors(rc);
        ddec := NumData;
      end
      else
      begin
        rc := SQLColAttribute(hStmt, i, SQL_COLUMN_LENGTH, @mbNumData, 32, msglen, NumData);
        //rc := SQLColAttribute(hStmt, i, SQL_COLUMN_LENGTH, pointer(mNumData), SQL_IS_INTEGER, msglen, NumData);
        ShowErrors(rc);
        dwidth := NumData;
        // added this to account for SqlServer allowing varchar(max) which ends up with a length of 0
        if((rc = 0) and (dwidth = 0))then
          if(fParent.GetDataType = dbSqlServer)then
            dwidth := 8000;
      end;
    end;
  end;
end;

function TODBCStatement.GetColDesc(field: string; var dtype, dwidth, ddec: integer): boolean;
begin
  Result := GetColDescEx(field, dtype, dwidth, ddec);
  if((dtype = integer('T')) or (dtype = integer('H')))then
    dtype := integer('D')
  else
    if(dtype = integer('F'))then
      dtype := integer('N');
end;

// same as above, but save the unmodifyed dtype in extype
function TODBCStatement.GetColDesc(field: string; var dtype, dwidth, ddec, extype: integer): boolean;
begin
  Result := GetColDescEx(field, dtype, dwidth, ddec);
  extype := dtype;
  if((dtype = integer('T')) or (dtype = integer('H')))then
    dtype := integer('D')
  else
    if(dtype = integer('F'))then
      dtype := integer('N');
end;

// Moves the record pointer according to the orientation selected, with the "amount"
// used for some.  Returns FALSE upon failure.  Bound variables will be updated with
// values for the newly pointed to record (row) 'automagically'.
function TODBCStatement.dbSkip(oType: TODBCOrientation; amount: integer): boolean;
var
  rc: integer;
  tmr: cardinal;
begin
  // for some reason, starting in Win2000, the buffers DO NOT clear
  // out the last row's value when field contains null data... !!!
  if(fBufferRec)then
    ClearFastRecBuff;

  // make sure we are 'allowed' to do the requested movement...
  case oType of
    oFirst,
    oLast,
    oAbsol:
      if(not fSqlFetchAbs)then
        oType := TODBCOrientation(-1);

    oPrior,
    oRelat:
      if(not fSqlFetchRel)then
        oType := TODBCOrientation(-1);

    //oNext:
    //  if(not fSqlFetchNext)then
    //    ??;
  end;

  tmr := GetTickCount();
  // at this point, I don't believe these  FetchScroll's
  // work with Oracle's ODBC... so better only use oNext?
  if(fNoFetchScroll)then
    rc := SQLFetch(hStmt)
  else
    case oType of
      // far as I can tell, this is the same as FetchScroll with FETCH_NEXT
      // but appears more compatible with different data sources, so use it here...
      // if you encounter a problem with this, put the FETCH_NEXT
      // back, and let me know, k?  Dave
      oNext:  rc := SQLFetch(hStmt);  // rc := SQLFetchScroll(hStmt, SQL_FETCH_NEXT, 0);

      oFirst: rc := SQLFetchScroll(hStmt, SQL_FETCH_FIRST, 0);
      oLast:  rc := SQLFetchScroll(hStmt, SQL_FETCH_LAST, 0);
      oAbsol: rc := SQLFetchScroll(hStmt, SQL_FETCH_ABSOLUTE, amount);

      oPrior: rc := SQLFetchScroll(hStmt, SQL_FETCH_PRIOR, 0);
      oRelat: rc := SQLFetchScroll(hStmt, SQL_FETCH_RELATIVE, amount);

      else
        rc := SQL_ERROR;
    end;

  TimeLog(GetTickCount() - tmr, 'dbSkip', fLastSQL);

  Result := ((rc = SQL_SUCCESS_WITH_INFO) or (rc = SQL_SUCCESS));
end;

// get the data in the column into the buffer pointed to by Data and
// return the length read...
function TODBCStatement.GetData(col: integer; dtyp: smallint; Data: pointer; dLen: integer = 0): integer;
var
  rc: integer;
begin
  Result := 0;
                                                     // -1 result when data is equal null
  rc := SQLGetData(hStmt, word(col + 1), dtyp, Data, dLen, Result);

  // this didn't work either...
//  if(rc = -1)then
//    if(fBufferRec)then
//      Result := GetDataWidth(col);

  // Should this field been extracted from already, we must refresh
  // the buffers before getting the data again.
  if(rc = SQL_NO_DATA)then
  begin
    if(fSqlPosRefresh)then
      if(SQLSetPos(hStmt, 1, SQL_REFRESH, SQL_LOCK_NO_CHANGE) <> SQL_ERROR)then
        {rc := }SQLGetData(hStmt, word(col + 1), dtyp, Data, dLen, Result);
  end;
end;

// get the actual size of the data in a column for the current row...
// NOTE:
//   the only way I could find to get the length of a blob was
//   to read in the data into a field, so I read it asking for a length of 0, just
//   to get the length, then we can allocate memory for the 'real' read...
function TODBCStatement.GetDataLen(col: integer; dt: TODBCDataTypes): integer;
var
  buf: byte;
begin
  if(dt <> dBlob)then
    Result := GetColWidth(col)
  else
    Result := GetData(col, SQL_CHAR, @buf);
end;

// determine if a given columns value is null...
//  where x=''  and  where x is null   will not always work the same
function TODBCStatement.IsNull(field: string): boolean;
var
  col: integer;
  buf: byte;
begin
  col := FindColName(field);
  if(col >= 0)then
    Result := (GetData(col, SQL_CHAR, @buf) = SQL_NULL_DATA)
  else
    Result := True;
end;

// Fetches a string value from current record for
// a given field column...
function TODBCStatement.fetch(col: integer): string;
var
  buf: TodBytes;
  dt: TODBCDataTypes;
  aLen: integer;
begin
  dt := GetColType(col);
  if((not fBufferRec) or (dt = dBlob))then
  begin
    // first, determine the length of this data...
    alen := GetDataLen(col, dt);
    if(aLen > 0)then
    begin
      SetLength(buf, aLen + 2);       // I like to have a 'little head room'!?
      aLen := GetData(col, SQL_CHAR, buf, aLen + 1);  // + 1 = space for null terminator...

      SetString(Result, PChar(buf), aLen);
      SetLength(buf, 0);
      buf := nil;
      Result := Trim(Result);
    end
    else
      Result := '';
  end
  else
    Result := ToString(fastRecBuff[col]);
end;

// Fetches a string value from current record for
// a given field column...
function TODBCStatement.bfetch(field: string; var buf: Pointer): integer;
var
  dt: TODBCDataTypes;
  aLen, col: integer;
begin
  Result := 0;
  col := FindColName(field);
  if(col >= 0)then
  begin
    dt := GetColType(col);
    // first, determine the length of this data...
    alen := GetDataLen(col, dt);
    if(aLen > 0)then
    begin
      buf := AllocMem(aLen + 2);
      Result := GetData(col, SQL_BINARY, buf, aLen);  // + 1 = space for null terminator...
    end;
  end;
end;

// Fetches a string value from current record for
// a given field name...
function TODBCStatement.fetch(field: string): string;
var
  col: integer;
begin
  col := FindColName(field);
  if(col >= 0)then
    Result := fetch(col)
  else
    Result := '';
end;

// rb Feb 29/2000, strip the '_' from the string on fetch
// used by BMO on Credit card upload
function TODBCStatement.fetchNu(field: string): string;
begin
  Result := StringReplace(fetch(field), '_', '-', [rfReplaceAll]);
end;

// Fetches an integer value from current record pointer spot for
// the specified field name...
function TODBCStatement.ifetch(field: string): integer;
var
  col: integer;
begin
  Result := 0;
  col := FindColName(field);
  if(col >= 0)then
    if(fBufferRec)then
      Result := StrToIntDef(ToString(fastRecBuff[col]), 0)
    else
      GetData(col, SQL_INTEGER, @Result);
end;

// this function is not in Delphi 5, so this makes sure we are compatible...
function myStrToFloatDef(const S: string; const Default: Extended): Extended;
begin
  if(not TextToFloat(PChar(S), Result, fvExtended))then
    Result := Default;
end;

// this function is not in Delphi 5, so this makes sure we are compatible...
function myStrToCurrencyDef(const S: string; const Default: Currency): Currency;
begin
  if(not TextToFloat(PChar(S), Result, fvCurrency))then
    Result := Default;
end;

// Fetches a double value from current record pointer spot for the specified
// field name.
function TODBCStatement.dfetch(field: string): double;
var
  col: integer;
begin
  Result := 0.0;
  col := FindColName(field);
  if(col >= 0)then
    if(fBufferRec)then
      Result := myStrToFloatDef(ToString(fastRecBuff[col]), 0.0)
    else
      GetData(col, SQL_DOUBLE, @Result);
end;

// Fetches a currency value from current record pointer spot
// for the specified field name.
function TODBCStatement.cfetch(field: string): currency;
var
  Temp: double;
  col: integer;
begin
  Result := 0.0;
  col := FindColName(field);
  if(col >= 0)then
    if(fBufferRec)then
      Result := myStrToCurrencyDef(ToString(fastRecBuff[col]), 0.0)
    else
    begin
      GetData(col, SQL_DOUBLE, @Temp);
      Result := Temp;
    end;
end;

// This will fetch dollar amounts as integers; otherwise, if you decide to use
// "dfetch", BE WARNED:  adding up or subtracting the values can lead to small
// 'in-equalities' (ie. 2.00 + 2.00 MAY NOT = 4.00 !!!!!)
//                      [probably due to the conversion from text to double used if fastRecBuf is on, "2.00" can in fact be 1.999999999999999 when made a number!]).
function TODBCStatement.mfetch(field: string): integer;
begin
  Result := Round(cfetch(field) * 100);
end;

// Fetches a boolean value from current record pointer spot for the specified
// field name.
function TODBCStatement.tfetch(field: string): boolean;
//var
//  col, junk, rc: integer;
begin
  // I don't believe Oracle supports this type, so if tfetch is used,
  // I will ass-u-me it's an integer field and a non zero value is true...
  Result := (ifetch(field) <> 0);

  (*
  Result := False;
  col := FindColName(field);
  if(col >= 0)then
    if(fBufferRec)then
      // I'M SURE THIS IS INCORRECT,
      // WHAT DOES A 'LOGICAL' LOOK LIKE AS A STRING?
      Result := (UpperCase(ToString(fastRecBuff[col])) = 'TRUE')
    else
      GetData(col, SQL_BIT, @Result);
  *)
end;

// Fetches a Date, Time or Date Time value from current record pointer spot for
// the specified field name.
function TODBCStatement.dtfetch(field: string): TDateTime;
var
  rs: string;
  dt: TODBCDataTypes;
  col: integer;
begin
  Result := Date;

  col := FindColName(field);
  if(col >= 0)then
  begin
    dt := GetColType(col);
    // If the data type for this field is different, don't even bother.
    if((dt = dDate) or (dt = dTime) or (dt = dDateTime))then
    begin
      rs := fetch(col);

      // Translate the  string result to Delphi's TDateTime.
      if(dt = dDate)then
        Result := BoundToDate(rs)
      else
        if(dt = dTime)then
          Result := BoundToTime(rs)
        else
          Result := BoundToDateTime(rs);
    end;
  end;
end;

function TODBCStatement.BindToField(field: string; variable: pointer): boolean;
{ The only way to write values to a table, is to bind the fields you wish to
  alter with a variable address you specify, then upon changing the value at
  that address and calling ReplaceRow, the values for the bound columns (fields)
  in the currently pointed to record is updates.  Each dbSkip called will also
  update the value in the variable with the value in the column of the
  currently pointed to record (so it has a dual purpose that can be used for
  quick, repeated reads for a result set!). All types of variables to be used
  for binding should be the same as the type in the column, except all string,
  data, time and datetime variables should by type ODBCString.
  NOTE:  Of course you can always use UPDATE or INSERT statements. }
var
  col: PColInfo;
  i, rc: integer;
begin
  Result := False;

  i := FindColName(field);
  if(i >= 0)then
  begin
    col := ColInfo.Items[i];
    case GetColType(i) of
      dChar   : rc := SQLBindCol(hStmt, word(i + 1), SQL_CHAR, variable, 255, @col^.BufSiz);
      dInt    : rc := SQLBindCol(hStmt, word(i + 1), SQL_INTEGER, variable, 4, @col^.BufSiz);
      dFloat  : rc := SQLBindCol(hStmt, word(i + 1), SQL_DOUBLE, variable, 255, @col^.BufSiz);
      dLogical: rc := SQLBindCol(hStmt, word(i + 1), SQL_BIT, variable, 1, @col^.BufSiz);
      dDate   : rc := SQLBindCol(hStmt, word(i + 1), SQL_CHAR, variable, 255, @col^.BufSiz);
      dTime   : rc := SQLBindCol(hStmt, word(i + 1), SQL_CHAR, variable, 255, @col^.BufSiz);
      dDateTime: rc := SQLBindCol(hStmt, word(i + 1), SQL_CHAR, variable, 255, @col^.BufSiz);
    else
      rc := 0;
    end;

    Result := ((rc = SQL_SUCCESS) or (rc = SQL_SUCCESS_WITH_INFO));
  end;
end;

function TODBCStatement.ReplaceRow: boolean;
{ Should the ODBC Class object be set for write access and columns have been
  bound to program variables, this will update the currently pointed to row
  with the values pointed to by bound variables. }
var
  rc: integer;
begin
  if(fSqlPosUpdate)then
  begin
    rc := SQLSetPos(hStmt, 1, SQL_UPDATE, SQL_LOCK_NO_CHANGE);
    Result := ((rc = SQL_SUCCESS) or (rc = SQL_SUCCESS_WITH_INFO));
  end
  else
    Result := False;
end;

// Because of the way ODBC stores dates and times, this will convert ODBCString
// values representing dates to TDateTime. }
function TODBCStatement.BoundToDate(d: string): TDateTime;
var
  Dat: TDateTime;
begin
{$ifdef GTR_50}
  Result := 0.0;
  if(TryEncodeDate(StrToIntDef(Copy(d, 1, 4), 1899), StrToIntDef(Copy(d, 6, 2), 12), StrToIntDef(Copy(d, 9, 2), 30), Dat))then
    Result := DateOf(Dat);
{$else}
 try
  Dat := EncodeDate(StrToIntDef(Copy(d, 1, 4), 1899), StrToIntDef(Copy(d, 6, 2), 12), StrToIntDef(Copy(d, 9, 2), 30));
  Result := Trunc(Dat);
 except
  Result := 0.0;
 end;
{$endif}
end;

{ Because of the way ODBC stores dates and times, this will convert ODBCString
  values representing times to TDateTime. }
function TODBCStatement.BoundToTime(t: string): TDateTime;
var
  Tim: TDateTime;
begin
{$ifdef GTR_50}
  Result := 0.0;
  if(TryEncodeTime(StrToIntDef(Copy(t, 1, 2), 0), StrToIntDef(Copy(t, 4, 2), 0), StrToIntDef(Copy(t, 7, 2), 0), 0, Tim))then
    Result := TimeOf(Tim);
{$else}
  try
    Tim := EncodeTime(StrToIntDef(Copy(t, 1, 2), 0), StrToIntDef(Copy(t, 4, 2), 0), StrToIntDef(Copy(t, 7, 2), 0), 0);
    Result := Frac(Tim);
  except
    Result := 0.0;
  end;
{$endif}
end;

// Because of the way ODBC stores dates and times, this will convert ODBCString
// values representing datetimes to TDateTime.
//   I DIDN'T LIKE THE FACT THAT EMPTY FIELDS (or any invalid DateTime)
//   ENDED UP NOW, CHANGED TO 0.0  Sept,2005  DKM
function TODBCStatement.BoundToDateTime(dt: string): TDateTime;
var
  Dat, Tim: TDateTime;
begin
{$ifdef GTR_50}
  Result := 0.0;
  if(TryEncodeDate(StrToIntDef(Copy(dt, 1, 4), 1899), StrToIntDef(Copy(dt, 6, 2), 12), StrToIntDef(Copy(dt, 9, 2), 30), Dat))then
    if(TryEncodeTime(StrToIntDef(Copy(dt, 12, 2), 0), StrToIntDef(Copy(dt, 15, 2), 0), StrToIntDef(Copy(dt, 18, 2), 0), 0, Tim))then
      Result := DateOf(Dat) + TimeOf(Tim);
{$else}
 try
  Dat := EncodeDate(StrToIntDef(Copy(dt, 1, 4), 1899), StrToIntDef(Copy(dt, 6, 2), 12), StrToIntDef(Copy(dt, 9, 2), 30));
  Tim := EncodeTime(StrToIntDef(Copy(dt, 12, 2), 0), StrToIntDef(Copy(dt, 15, 2), 0), StrToIntDef(Copy(dt, 18, 2), 0), 0);
  Result := Trunc(Dat) + Frac(Tim);
 except
  Result := 0.0;
 end;
{$endif}
end;

// Because of the way ODBC stores dates and times, this will convert TDateTime
// values representing dates to ODBCString.
function TODBCStatement.DateToBound(d: TDateTime): ODBCString;
begin
  StrPCopy(PChar(@Result), FormatDateTime('yyyy-mm-dd', d));
end;

// Because of the way ODBC stores dates and times, this will convert TDateTime
// values representing times to ODBCString.
function TODBCStatement.TimeToBound(t: TDateTime): ODBCString;
begin
  StrPCopy(PChar(@Result), FormatDateTime('hh:nn:ss', t));
end;

// Because of the way ODBC stores dates and times, this will convert TDateTime
// values representing datetimes to ODBCString.
function TODBCStatement.DateTimeToBound(dt: TDateTime): ODBCString;
begin
  StrPCopy(PChar(@Result), FormatDateTime('yyyy-mm-dd hh:nn:ss', dt));
end;

function TODBCStatement.GetODBCError: string;
var
  errMsg: string;
  i: integer;
begin
  i := 1;
  Result := '';
  while(AnyODBCErrors(i, errMsg))do
  begin
    if(Length(Result) > 0)then
      Result := Result + #13#10;
    Result := Result + errMsg;
    Inc(i);
  end;
end;

procedure TODBCStatement.ShowErrors(rc: integer);
var
  errMsg: string;
  i: integer;
begin
  if(fParent.DoShowErrors)then
    if((rc <> SQL_SUCCESS) and (rc <> SQL_SUCCESS_WITH_INFO))then
    begin
      i := 1;
      while(AnyODBCErrors(i, errMsg))do
      begin
{$ifndef NO_MESSLOG}
        if(Assigned(fParent.fMess))then
          fParent.fMess.Flush(errMsg)
        else
{$endif}
          MessageDlg('Stmt Error: ' + errMsg, mtError, [mbOk], 0);
        Inc(i);
      end;
    end;
end;

function TODBCStatement.CheckErr: integer;
var
  Msg, State: string;
begin
  Result := GetErrMsg(1, Msg, State);
end;

function TODBCStatement.GetErrMsg(ndx: smallint; var eMess, eState: string): integer;
var
  msg: array [0..SQL_MAX_MESSAGE_LENGTH] of char;
  sqlState: array [0..8] of char;
  nativeError: integer;
  msgLength: smallint;
begin
  eMess := '';
  eState := '';
  Result := SQLGetDiagRec(SQL_HANDLE_STMT, hStmt, ndx, @sqlState, nativeError, @msg, SQL_MAX_MESSAGE_LENGTH, msgLength);
  if((Result = SQL_SUCCESS) or (Result = SQL_SUCCESS_WITH_INFO))then
  begin
    SetString(eState, sqlState, StrLen(sqlState));

    if(eState = 'HY000')then
    begin
      Result := SQLGetDiagField(SQL_HANDLE_STMT, hStmt, ndx, SQL_DIAG_MESSAGE_TEXT, @msg, SQL_MAX_MESSAGE_LENGTH, msglength);
      if((Result = SQL_SUCCESS) or (Result = SQL_SUCCESS_WITH_INFO))then
        SetString(eMess, msg, StrLen(msg));
    end
    else
      //      if(eState = '01S01')then    // ignoring 'Error in row' for now?
      //        Result := SQL_NEED_DATA
      //      else
      SetString(eMess, msg, StrLen(msg));

    fParent.CheckConnFail(eState, eMess);
  end;
end;

// For returning an error message at index "ndx" from the Statement handle.
// Note: "ndx" must count up from 1, and keep going till this function returns
//       FALSE.
function TODBCStatement.AnyODBCErrors(ndx: smallint; var errMsg: string): boolean;
var
  SQLState: string;
  rc: integer;
begin
  rc := GetErrMsg(ndx, errMsg, SQLState);
  // Test to see if there are any "Diag" records to show.
  Result := ((rc = SQL_SUCCESS) or (rc = SQL_SUCCESS_WITH_INFO));
  if(Result)then
  begin
    // Ignore if error is:  "Driver doesn't support...".
    if(SQLState = 'IM001')then
      Result := False
    else
      errMsg := errMsg + '. State:  ' + SQLState;
  end;
end;

constructor TODBC.Create;
var
  rc: integer;
begin
  inherited Create;

  hDBC := nil;
  if(SetODBCProcs)then
  begin
    // DLL is now succefully loaded, allocate the ODBC environment handle.
    rc := SQLAllocHandle(SQL_HANDLE_ENV, SQL_NULL_HENV, hEnv);
    if((rc = SQL_SUCCESS) or (rc = SQL_SUCCESS_WITH_INFO))then
    begin
      // Ensure that ODBC engine is to the ODBC version 3.x standard at least.
      rc := SQLSetEnvAttr(hEnv, SQL_ATTR_ODBC_VERSION, pointer(SQL_OV_ODBC3), 0);
      if((rc = SQL_SUCCESS) or (rc = SQL_SUCCESS_WITH_INFO))then
      begin
        fShowErrors := False;
        fDSConnected := False;
        fConnectionFailure := False;
        fConnString := '';
        fDatabType := dbTypeUnknown;

        fAllowSelectEmptySet := False;
{$ifndef NO_MESSLOG}
        fMess := nil;
{$endif}
{$ifdef ISASERVICE}
  {$ifndef NOTHREAD}
        fAppThread := nil;
  {$endif}
{$endif}
      end;
    end;
  end;
end;

// Legacy function
// can now just do  ODBCMain := TODBC.Create;
// and it will do the above create function
//
// Actually creates the ODBC Class object...   called from CreateODBC;
constructor TODBC.Create(environmentHandle: SQLHENV);
begin
  inherited Create;

  hEnv := environmentHandle;
  fShowErrors := False;
  fDSConnected := False;
  fConnectionFailure := False;
  fConnString := '';
  fDatabType := dbTypeUnknown;

  fAllowSelectEmptySet := False;       // The old behaviour will be the default mode, added by KRS 3/1/2006.
{$ifndef NO_MESSLOG}
  fMess := nil;
{$endif}
{$ifdef ISASERVICE}
 {$ifndef NOTHREAD}
  fAppThread := nil;
 {$endif}
{$endif}
  DTmr := 0;
end;

{$ifdef LIKE_UEXPRESS}
constructor TCBase.Create(HostName, Database, User_name, Password: string);
begin
  inherited Create;    // must call the ancestors create

  ConnectToDataSource(HostName, User_name, Password, Database);
end;

function TSQLStatement.skip: boolean;
begin
  Result := dbskip;
end;
{$endif}

// Destroys the ODBC object by freeing all of the ODBC32.DLL handles,
// disconnecting from the data source and destroys the result field list if any.
destructor TODBC.Destroy;
begin
  // sometimes, we cannot disconnect without doing this first,
  // or we get an 'Invalid Transaction state'...
  //CommitSQL(True);
  
  { This will free all Statement handles and the current Data Source Handle.
    RB, Sept. 6, 2000, moved this, think it was causing a problem being called
    before "FreeStatement".  KRS 12/4/2000, he's absolutely right. }
  DisconnectFromDataSource;

  // Delete the now empty statement array.
  Finalize(Stmts);

  // Free the ODBC environment handle.
  ShowErrors(SQLFreeHandle(SQL_HANDLE_ENV, hEnv));
  hEnv := nil;

  // Free the ODBC32.DLL library.
  FreeLibrary(hODBC32DLL);

  inherited Destroy;
end;

function TODBC.SetDataType(dbType: TdbType): TdbType;
begin
  Result := fDatabType;
  fDatabType := dbType;
end;

function TODBC.GetDataType: TdbType;
begin
  Result := fDatabType;
end;

// Populates the TStringList "List" with all of the registered
// driver names (or descriptions in the SAME ORDER) one the system for ODBC.
procedure TODBC.GetDriversList(drvrNotDesc: boolean; list: TStringList);
var
  sDriver, sDesc: string;
  driver, desc: ODBCString;
  rc: integer;
  lDrvr, lDesc: smallint;
begin
  list.Clear;

  lDrvr := 0;
  lDesc := 0;
  rc := SQLDrivers(hEnv, SQL_FETCH_FIRST, @driver, 255, lDrvr, @desc, 255, lDesc);
  while((rc = SQL_SUCCESS) or (rc = SQL_SUCCESS_WITH_INFO))do
  begin
    if(drvrNotDesc)then
    begin
      SetString(sDriver, PChar(@driver), lDrvr);
      list.Add(sDriver);
    end
    else
    begin
      SetString(sDesc, PChar(@desc), lDesc);
      list.Add(sDesc);
    end;
    lDrvr := 0;
    lDesc := 0;
    rc := SQLDrivers(hEnv, SQL_FETCH_NEXT, @driver, 255, lDrvr, @desc, 255, lDesc);
  end;
end;

// Populates the TStringList "List" with all of the registered
// Data Sources (or descriptions in the SAME ORDER) one the system for ODBC.
procedure TODBC.GetDataSourcesList(datSrcNotDesc: boolean; list: TStringList; DSNtoDvr: PChar);
var
  sSource, sDesc: string;
  source, desc: ODBCString;
  rc: integer;
  lSrc, lDesc: smallint;
begin
  if(list <> nil)then
    list.Clear;

  lSrc := 0;
  lDesc := 0;
  rc := SQLDataSources(hEnv, SQL_FETCH_FIRST, @source, 255, lSrc, @desc, 255, lDesc);
  while((rc = SQL_SUCCESS) or (rc = SQL_SUCCESS_WITH_INFO))do
  begin
    if(DSNtoDvr <> nil)then
    begin
      SetString(sSource, PChar(@source), lSrc);
      if(sSource = DSNtoDvr)then
      begin
        StrCopy(DSNtoDvr, PChar(@desc));
        break;
      end;
    end
    else
      if(datSrcNotDesc)then
      begin
        SetString(sSource, PChar(@source), lSrc);
        list.Add(sSource);
      end
      else
      begin
        SetString(sDesc, PChar(@desc), lDesc);
        list.Add(sDesc);
      end;

    lSrc := 0;
    lDesc := 0;
    rc := SQLDataSources(hEnv, SQL_FETCH_NEXT, @source, 255, lSrc, @desc, 255, lDesc);
  end;
end;

// build the setup string passed to SQLDriverConnect...
procedure AddinSetting(var str: string; setting, value: string);
begin
  if(Length(value) > 0)then
  begin
    if(Length(str) > 0)then
      str := str + ';';
    str := str + setting + '=' + value;
  end;
end;

// this is to get some 'auto detect' of the database type, but it may need
// adjusting in future, or you can override it with a call to SetDataType
procedure TODBC.AutodbType(s: string);
begin
  s := LowerCase(s);
  if(Pos('oracle', s) <> 0)then
    fDatabType := dbTypeOracle
  else
    if(Pos('firebird', s) <> 0)then
      fDatabType := dbTypeFireBird
    else
      if(Pos('mysql', s) <> 0)then
        fDatabType := dbTypeMySQL
      else
        if((Pos('sql server', s) <> 0) or (Pos('sql native client', s) <> 0))then
          fDatabType := dbSqlServer
        else
          if(Pos('codebase', s) <> 0)then
            fDatabType := dbCodeBase
          else
            if((Pos('foxpro', s) <> 0) or (Pos('dbf', s) <> 0))then
              fDatabType := dbTypeDBF
            else
              if(Pos('access', s) <> 0)then
                fDatabType := dbAccess
              else
                if(Pos('excel', s) <> 0)then
                  fDatabType := dbExcel;
end;

// as of yet, I haven't found any datasource that returns dead under
// ANY circumstances...
//  (this may be a fairly new attribute that is not widely supported yet)
function TODBC.ConnectionDead: boolean;
var
  lInfo: Cardinal;
  lRet: integer;
begin
  if(fConnectionFailure)then      // this is set in AnyODBCError if state is 08xxx
    Result := True
  else
  begin
    ShowErrors(SQLGetConnectAttr(hDBC, SQL_ATTR_CONNECTION_DEAD, @lInfo, SizeOf(lInfo), lRet));
    Result := (lInfo = SQL_CD_TRUE);
  end;
end;

// this does a connection with as few ODBC calls as possible...
// NOTE:  DOES NOT WORK WITH ALL DATASOURCES!!!  CURRENTLY ONLY
//        TESTED AND WORKING WITH FIREBIRD and MSACCESS
function TODBC.FastConnect(dataSource, login, password, catalog: string; Driver: string = ''; AddSettings: string = ''): boolean;
var
  name: array[0..1023] of char;
  ConnParams, DBTag: string;
  rc: integer;
  sRet: smallint;
begin
  if(not fDSConnected)then
  begin
    // Allocate the ODBC connection handle.
    rc := SQLAllocHandle(SQL_HANDLE_DBC, hEnv, hDBC);
    if((rc = SQL_SUCCESS) or (rc = SQL_SUCCESS_WITH_INFO))then
    begin
      // both of these don't seem to make much, if any, difference?
      ShowErrors(SQLSetConnectAttr(hDBC, SQL_ATTR_CONNECTION_TIMEOUT, pointer(2), 0));
//      ShowErrors(SQLSetConnectAttr(hDBC, SQL_LOGIN_TIMEOUT, pointer(1), 0));
(*

      lRet := 0;
      ConnParams := 'DRIVER={Firebird/InterBase(r) driver}';
      rc := SQLBrowseConnect(hDBC, PChar(ConnParams), SQL_NTS, name, SizeOf(name), lRet);
      SetString(fConnString, name, lRet);
      ShowMessage(fConnString);
      if(rc = SQL_NEED_DATA)then
      begin
        lRet := 0;
        ConnParams := 'UID=' + Login + ';PWD=' + Password;
        rc := SQLBrowseConnect(hDBC, PChar(ConnParams), SQL_NTS, name, SizeOf(name), lRet);
        SetString(fConnString, name, lRet);
        ShowMessage(fConnString);
        if(rc = SQL_NEED_DATA)then
        begin
          lRet := 0;
          ConnParams := 'DBNAME=209.161.244.74:c:\program files\firebird\firebird_1_5\MAS.fdb';
          ConnParams := 'DBNAME=c:\program files\firebird\firebird_1_5\MAS.fdb';
          ConnParams := 'DBNAME=209.161.244.74:';
          rc := SQLBrowseConnect(hDBC, PChar(ConnParams), SQL_NTS, name, SizeOf(name), lRet);
          if(rc = SQL_NEED_DATA)then
          begin
            SetString(fConnString, name, lRet);
            ShowMessage(fConnString);
          end;
        end;
      end;
*)
      // MUST have this for Access (maybe others, but not FireBird or Oracle)...
      if(Pos('access', LowerCase(Driver)) <> 0)then
        if(Length(Catalog) > 0)then
          ShowErrors(SQLSetConnectAttr(hDBC, SQL_ATTR_CURRENT_CATALOG, PChar(Catalog), SQL_NTS));

      if(Length(DataSource) > 0)then
        // Actually connect to the data source (DSN) with login and password.
        rc := SQLConnect(hDBC, PChar(DataSource), SQL_NTS, PChar(Login), SQL_NTS, PChar(Password), SQL_NTS)
      else
        rc := SQL_ERROR;     // if no DSN name passed, force using SQLDriverConnect below...

      // when the above fails, it may be due to incomplete information in the Data Source,
      // so try again with the following and it *may* get you in!...
      // actually, you don't need a DSN setup, if you use this with DRIVER set!?!
      if((rc <> SQL_SUCCESS) and (rc <> SQL_SUCCESS_WITH_INFO))then
      begin
        // you could try putting more options in this string???
        ConnParams := '';
        AddinSetting(ConnParams, 'DSN', dataSource);
        AddInSetting(ConnParams, 'DRIVER', '{' + Driver + '}');

        if(Pos('oracle', LowerCase(Driver)) <> 0)then
          DBTag := 'DBQ'
        else
          if(Pos('foxpro', LowerCase(Driver)) <> 0)then
            DBTag := 'SourceDB'
          else
            DBTag := 'Database';

        AddinSetting(ConnParams, DBTag, catalog);
        if(Length(Login) > 0)then
          AddInSetting(ConnParams, 'UID', Login);
        if(Length(Password) > 0)then
          AddInSetting(ConnParams, 'PWD', Password);

        // any 'custom' settings passed in this string...
        //  (ie. or 'SourceType=dbf' for FOX)
        if(Length(AddSettings) > 0)then
        begin
          if(Length(ConnParams) > 0)then
            ConnParams := ConnParams + ';';
          ConnParams := ConnParams + AddSettings;
        end;

        sRet := 0;                                                                                      // pass window handle in 2nd param and use below option to get dialog prompting when data is incomplete...
        rc := SQLDriverConnect(hDBC, HWND(nil), PChar(ConnParams), SQL_NTS, @name, 1023, sRet, SQL_DRIVER_NOPROMPT); // SQL_DRIVER_COMPLETE);
        SetString(fConnString, name, sRet);
      end
      else
      begin
        // get driver name from DSN name for AutodbType()
        StrPCopy(name, DataSource);
        GetDataSourcesList(False, nil, name);
        SetString(fConnString, name, StrLen(name));
      end;

      if((rc = SQL_SUCCESS) or (rc = SQL_SUCCESS_WITH_INFO))then
      begin
        // 'try' to figure out the database type...
        AutodbType(fConnString);
        fConnectionFailure := False;
        fDSConnected := True;
      end
      else
      begin     // If we failed to connect, free handle and return FALSE.
        ShowErrors(rc);
        SQLFreeHandle(SQL_HANDLE_DBC, hDBC);
        hDBC := nil;
      end;
    end;
  end;
  Result := fDSConnected;
end;

//  get the DSN Info ; separated, if filesource
function TODBC.GetDSNInfo(filePath: string): string;
var
  list: TStringList;
  ini: TIniFile;
  val: string;
  ndx: integer;
begin
  Result := '';
  list := TStringList.Create;
  ini := TIniFile.Create(filePath);
  ini.ReadSection('ODBC', list);  // get a list of all keys in ODBC section...
  for ndx := 0 to list.Count - 1 do
  begin
    val := ini.ReadString('ODBC', list.Strings[ndx], '');
    // if the driver value has spaces, put {} around it
    // (that seems to be what ODBC likes?!)
    if(Pos(' ', val) > 0)then
      if(UpperCase(list.Strings[ndx]) = 'DRIVER')then
        val := '{' + val + '}';
//      else
//        val := '"' + val + '"';

    // and add the key=value into the result string...
    if(Length(Result) > 0)then
      Result := Result + ';';
    Result := Result + list.Strings[ndx] + '=' +  val;
  end;
  ini.Free;
  list.Free;
end;

// Connects to a specified Data Source with a login name, password and initial
// catalog (even though tables can for the most part be selected from any
// different catalog later, it's here to support the largest variety of drivers).
function TODBC.ConnectToDataSource(dataSource, login, password, catalog: string;
                                    commitOff: boolean = False; Driver: string = ''; AddSettings: string = ''): boolean;
var
  name: array[0..1023] of char;
  sInfo: array [0..12] of char;
  ConnParams, DBTag: string;
  lInfo: Cardinal;
  iInfo, cAtt: word;
  rc: integer;
  sRet: smallint;
begin
  // We cannot connect to a new Data Source until the old one is freed, to
  // prevent memory leaks, by KRS 12/4/2000.
  if(not fDSConnected)then
  begin
    // Allocate the ODBC connection handle.
    rc := SQLAllocHandle(SQL_HANDLE_DBC, hEnv, hDBC);
    if((rc = SQL_SUCCESS) or (rc = SQL_SUCCESS_WITH_INFO))then
    begin
      // Set default login timeout
     //(* THIS AIN'T WORKIN'  ?????
      ShowErrors(SQLSetConnectAttr(hDBC, SQL_LOGIN_TIMEOUT, pointer(5), 0));
//    ShowErrors(SQLSetConnectAttr(hDBC, SQL_ATTR_CONNECTION_TIMEOUT, pointer(5), 0));
     //*)

     (*  Doesn't appear to work with Oracle (what else is new?) ...
      ShowErrors(SQLSetConnectAttr(hDBC, SQL_ATTR_ASYNC_ENABLE, pointer(SQL_ASYNC_ENABLE_ON), 0));
     *)

      AutoDBType(Driver);

      // Set default catalog.
      // rb Dec 1/03, this crashes Pervasive!!!  I guess they don't support catalogs!! (or was it that the catalog was blank?)
      // from now on if catalog is blank, then don't try and set it
      // CodeBase has some problem when you set this, not that you can't, but if you
      // do, you have to do something 'special' with the DBQ= and I never figured that out yet!?
      if(Length(Catalog) > 0)then
        if(fDatabType <> dbCodeBase)then
          ShowErrors(SQLSetConnectAttr(hDBC, SQL_ATTR_CURRENT_CATALOG, PChar(Catalog), SQL_NTS));

      // this is supposed to be the default, but had to do it for an INFORMIX database
      // don't show errors here as you may see 'not implemented' errors, and who cares!?
      // we do this for each statement handle as well...
      SQLSetConnectAttr(hDBC, SQL_ATTR_METADATA_ID, pointer(SQL_FALSE), SQL_IS_UINTEGER);

      if(commitOff)then
        // need this set off for commit / rollback to work... (in Oracle, at least?)
        ShowErrors(SQLSetConnectAttr(hDBC, SQL_ATTR_AUTOCOMMIT, pointer(SQL_AUTOCOMMIT_OFF), SQL_IS_INTEGER));

      (* Didn't do a thing for this problem, and don't want to make a 'new' one!
      ShowErrors(SQLSetConnectAttr(hDBC, SQL_ATTR_ACCESS_MODE, pointer(SQL_MODE_READ_WRITE), SQL_IS_INTEGER));
      *)
      //SQLSetConnectAttr(hDBC, 1224{SQL_COPT_SS_MARS_ENABLED}, pointer(1){SQL_MARS_ENABLED_YES}, SQL_IS_UINTEGER);

      ShowErrors(SQLSetConnectAttr(hDBC, SQL_ATTR_ODBC_CURSORS, pointer(SQL_CUR_USE_IF_NEEDED), SQL_IS_INTEGER));

      if(Length(DataSource) > 0)then
        // Actually connect to the data source (DSN) with login and password.
        rc := SQLConnect(hDBC, PChar(DataSource), SQL_NTS, PChar(Login), SQL_NTS, PChar(Password), SQL_NTS)
      else
        rc := SQL_ERROR;     // if no DSN name passed, force using SQLDriverConnect below...

      // when the above fails, it may be due to incomplete information in the Data Source,
      // so try again with the following and it *may* get you in!...
      // actually, you don't need a DSN setup, if you use this with DRIVER set!?!
      if((rc <> SQL_SUCCESS) and (rc <> SQL_SUCCESS_WITH_INFO))then
      begin
        // you could try putting more options in this string???
        ConnParams := '';

        // maybe they are using a file dsn?
        if(Length(DataSource) > 0)then
          if(FileExists(DataSource))then
            ConnParams := GetDSNInfo(DataSource);

        if(Length(ConnParams) = 0)then
        begin
          AddinSetting(ConnParams, 'DSN', dataSource);
          AddInSetting(ConnParams, 'DRIVER', '{'+Driver+'}');

          case fDatabType of
            dbTypeOracle,
            dbCodeBase:
              DBTag := 'DBQ';

            dbTypeDBF:
              DBTag := 'SourceDB';

            else
              DBTag := 'Database';
          end;

          AddinSetting(ConnParams, DBTag, catalog);
        end;

        if(Length(Login) > 0)then
          AddInSetting(ConnParams, 'UID', Login);
        if(Length(Password) > 0)then
          AddInSetting(ConnParams, 'PWD', Password);

        // any 'custom' settings passed in this string...
        //  (ie. or 'SourceType=dbf' for FOX)
        if(Length(AddSettings) > 0)then
        begin
          if(Length(ConnParams) > 0)then
            ConnParams := ConnParams + ';';
          ConnParams := ConnParams + AddSettings;
        end;

        sRet := 0;                                                                                      // pass window handle in 2nd param and use below option to get dialog prompting when data is incomplete...
        rc := SQLDriverConnect(hDBC, HWND(nil), PChar(ConnParams), SQL_NTS, @name, 1023, sRet, SQL_DRIVER_NOPROMPT); // SQL_DRIVER_COMPLETE);
        SetString(fConnString, name, sRet);
      end
      else
      begin
        // get driver name from DSN name for AutodbType()
        StrPCopy(name, DataSource);
        GetDataSourcesList(False, nil, name);
        SetString(fConnString, name, StrLen(name));
      end;

      if((rc = SQL_SUCCESS) or (rc = SQL_SUCCESS_WITH_INFO))then
      begin
        // 'try' to figure out the database type...
        AutodbType(fConnString);

        //ShowErrors(SQLGetInfo(hDBC, SQL_DRIVER_ODBC_VER, @sInfo, 10, sRet));

        // Obtain some basic information on how catalogs are to appear in SQL statments.
        sRet := 0;
        iInfo := 0;
        ShowErrors(SQLGetInfo(hDBC, SQL_FILE_USAGE, @iInfo, 2, sRet));
        case iInfo of
          SQL_FILE_NOT_SUPPORTED:
            fDSFileUsage := fuFileNotSup;
          SQL_FILE_TABLE:
            fDSFileUsage := fuFileTable;
          SQL_FILE_CATALOG:
            fDSFileUsage := fuFileCatalog;
        end;

        (*  Could use this to set the name length (currently is [0..128] in the TColInfo record)
        sRet := 0;
        iInfo := 0;
        ShowErrors(SQLGetInfo(hDBC, SQL_MAX_COLUMN_NAME_LEN, @iInfo, SizeOf(iInfo), sRet));
        *)

        sRet := 0;
        iInfo := 0;
        ShowErrors(SQLGetInfo(hDBC, SQL_CATALOG_LOCATION, @iInfo, SizeOf(iInfo), sRet));
        fCatInFront := not (iInfo = SQL_CL_END);

        sRet := 0;
        sInfo[0] := #0;
        ShowErrors(SQLGetInfo(hDBC, SQL_CATALOG_NAME_SEPARATOR, @sInfo, 5, sRet));
        fCatSep := sInfo;

        // if these 2 bits aren't on, there is a problem when
        // fetching from a column that has a lower number than
        // that of a previous fetch (as Oracle seems to have!), so buffer each 'record'...
        // (as of the latest patch and ODBC Driver, Oracle now seems to be ok fetching in any order)
        sRet := 2;
        lInfo := 0;
        ShowErrors(SQLGetInfo(hDBC, SQL_GETDATA_EXTENSIONS, @lInfo, SizeOf(lInfo), sRet));
        fBufferRecs := not ((lInfo and (SQL_GD_ANY_ORDER or SQL_GD_ANY_COLUMN)) =
                               (SQL_GD_ANY_ORDER or SQL_GD_ANY_COLUMN));


        if(not fBufferRecs)then
        begin
          // we can retrieve any column in any order, BUT
          // can we use SetPos to retrieve the same column twice?
          // if not, we will need to buffer recs...
          ShowErrors(SQLGetInfo(hDBC, SQL_SCROLL_OPTIONS, @lInfo, SizeOf(lInfo), sRet));

          // have to figure out what kind of cursor we will be using:
          if((lInfo and SQL_SO_DYNAMIC) <> SQL_SO_DYNAMIC)then
          begin
            cAtt := SQL_FORWARD_ONLY_CURSOR_ATTRIBUTES1;
            if((lInfo and SQL_SO_KEYSET_DRIVEN) = SQL_SO_KEYSET_DRIVEN)then
              if((fDatabType <> dbTypeOracle) and (fDatabType <> dbTypeMySQL))then
                cAtt := SQL_KEYSET_CURSOR_ATTRIBUTES1
          end
          else
            cAtt := SQL_DYNAMIC_CURSOR_ATTRIBUTES1;

          ShowErrors(SQLGetInfo(hDBC, cAtt, @lInfo, SizeOf(lInfo), sRet));
          fBufferRecs := ((lInfo and SQL_CA1_POS_REFRESH) <> SQL_CA1_POS_REFRESH);
        end;

        // Success, return TRUE.
        fDSConnected := True;
        fConnectionFailure := False;
      end
      else
      begin                          // If we failed to connect, free handle and return FALSE.
        fDatabType := dbTypeUnknown;
        ShowErrors(rc);
        SQLFreeHandle(SQL_HANDLE_DBC, hDBC);
        hDBC := nil;
      end;
    end;
  end;
  Result := fDSConnected;
end;

//
// used by DLL's for Re-Connecting to the ODBC source
// rb Nov 16/04
//
function TODBC.ReConnect(connectionString: string): boolean;
var
  outp: array[0..1023] of char;
  loutp: smallint;
  rc: integer;
begin
  rc := SQLAllocHandle(SQL_HANDLE_DBC, hEnv, hDBC);
  if((rc = SQL_SUCCESS) or (rc = SQL_SUCCESS_WITH_INFO))then
    rc := SQLDriverConnect(hDBC, HWND(nil), PChar(ConnectionString),
                           SQL_NTS, @outp, 1023, loutp, SQL_DRIVER_NOPROMPT);
  ShowErrors(rc);
  Result := (rc = SQL_SUCCESS);
end;

// there is a 'problem' with the FoxPro driver,
// once you do a select on a table, that table is
// left open (no matter WHAT you do) and can't be
// opened exclusivly by anyone else, even after the
// statement is completely closed.  Only solution found
// so far is to disconnect and reconnect, so here we go!

// disconnect and reconnect to ODBC
//  Used in FoxPro, to free up the tables, otherwise they
//  are 'in use' and cannot be used 'exclusively' by any other process...
//
// However, this can be slowwwww, so you might want to find a better way!?!?
procedure TODBC.Release;
var
  outp: array[0..1023] of char;
  i: integer;
  loutp: smallint;
begin
  if(fDSConnected)then
  begin
    for i := Low(Stmts) to High(stmts) do
      if(Stmts[i] <> nil)then
        Stmts[i].CloseCursor;

    // make sure we don't have an 'Invalid Transaction State'...
    CommitSQL(True);

    ShowErrors(SQLDisconnect(hDBC));
    ShowErrors(SQLDriverConnect(hDBC, HWND(nil), PChar(fConnString), SQL_NTS, @outp, 1023, loutp, SQL_DRIVER_NOPROMPT));
    fConnectionFailure := False;
    for i := Low(Stmts) to High(Stmts) do
      if(Stmts[i] <> nil)then
        Stmts[i].OpenCursor(True);
  end;
end;

// This will first free any open Statement handles, and then free the current
// Data Source Handle for TODBC.  Added by KRS 12/4/2000.
procedure TODBC.DisconnectFromDataSource;
var
  i: integer;
begin
  // Free all of the left over statements.
  for i := Low(Stmts) to High(stmts) do
    if(Stmts[i] <> nil)then
      Stmts[i].Free;

  if(fDSConnected)then
  begin
    CommitSQL(True);

    ShowErrors(SQLDisconnect(hDBC));
    ShowErrors(SQLFreeHandle(SQL_HANDLE_DBC, hDBC));
    hDBC := nil;

    fDSConnected := False;
    fConnectionFailure := False;
  end;
end;

// When 'AutoCommit' is turned off, this function MUST be called with either TRUE
// to commit the transactions, or FALSE to rollback the transactions, or the
// transactions will be 'lost' ...
//   Note:  DO NOT USE RunSQL() with 'COMMIT' or 'ROLLBACK' as the command!
function TODBC.CommitSQL(Commit: boolean): boolean;
var
  rc: integer;
begin
  if(fDSConnected)then
  begin
    if(Commit)then
      rc := SQLEndTran(SQL_HANDLE_DBC, hDBC, SQL_COMMIT)
    else
      rc := SQLEndTran(SQL_HANDLE_DBC, hDBC, SQL_ROLLBACK);
  end
  else
    rc := SQL_SUCCESS;
  Result := ((rc = SQL_SUCCESS) or (rc = SQL_SUCCESS_WITH_INFO));
end;

// Legacy function...
//  CAN NOW JUST USE:  TODBCStatement.Create(ODBCMain, readonly)
//
// Create a new statement instance...
function TODBC.NewStatement(readOnly: boolean; supportOracle: boolean = True): TODBCStatement;
var
  i: integer;
begin
  if(supportOracle)then
    fDatabType := dbTypeOracle;

  // Create the new statement.
  Result := TODBCStatement.Create(Self, fBufferRecs, readOnly);

  if(Result.ConnectStatus)then           // Can we create the statement?
  begin
    i := Low(Stmts);
    while(i <= High(Stmts))do
    begin
      if(Stmts[i] = nil)then
        break;
      Inc(i);
    end;
    if(i > High(Stmts))then                   // See if the current statement array is full.
      SetLength(Stmts, Length(Stmts) + 1);    // Increase the capacity for the statement array by one.
    Stmts[i] := Result;
  end
  else
  begin
      // Eliminate the useless statement.
    Result.Free;
    Result := nil;
  end;
end;

// Legacy function
//  CAN NOW JUST USE:  TODBCStatement.Free;
//
// Free the memory associated with the statement object 'theStmt'
procedure TODBC.FreeStatement(theStmt: TODBCStatement);
begin
  // This statement is already toast, so don't bother.  Added by KRS 12/5/2000.
  if(theStmt <> nil)then
    theStmt.Free;
end;

function TODBC.GetODBCError: string;
var
  errMsg: string;
  i: integer;
begin
  i := 1;
  Result := '';
  while(AnyODBCErrors(i, errMsg))do
  begin
    if(Length(Result) > 0)then
      Result := Result + #13#10;
    Result := Result + errMsg;
    Inc(i);
  end;
end;

procedure TODBC.ShowErrors(rc: integer);
var
  errMsg: string;
  i: integer;
begin
  if(fShowErrors)then
    if((rc <> SQL_SUCCESS) and (rc <> SQL_SUCCESS_WITH_INFO))then
    begin
      i := 1;
      while(AnyODBCErrors(i, errMsg))do
      begin
{$ifndef NO_MESSLOG}
        if(Assigned(fMess))then
          fMess.Flush(errMsg)
        else
{$endif}
          MessageDlg('ODBC Error: ' + errMsg, mtError, [mbOk], 0);
        Inc(i);
      end;
    end;
end;

procedure TODBC.CheckConnFail(eState, eMess: string);
var
  i: integer;
begin
  // if the flag is already on, why bother!?
  if(not fConnectionFailure)then
  begin
    for i := Low(ConnFailStates) to High(ConnFailStates) do
      if(eState = ConnFailStates[i])then
      begin
        fConnectionFailure := True;
        break;
      end;

    if(not fConnectionFailure)then
      for i := Low(ConnFailMsgs) to High(ConnFailMsgs) do
        if(Pos(ConnFailMsgs[i], UpperCase(eMess)) > 0)then
        begin
          fConnectionFailure := True;
          break;
        end;
  end;
end;

  // For returning an error message at index "ndx" from the Data Source handle.
// Note:  "ndx" must count up from 1, and keep going till this function returns
//        FALSE.
function TODBC.AnyODBCErrors(ndx: smallint; var errMsg: string): boolean;
var
  msg: array [0..SQL_MAX_MESSAGE_LENGTH] of char;
  sqlState: array [0..8] of char;
  eState: string;
  msgLength: smallint;
  rc: integer;
  nativeError: integer;
begin
  errMsg := '';
  msgLength := 0;
  // Get the status records.

  if(hDBC <> nil)then
  begin
    rc := SQLGetDiagRec(SQL_HANDLE_DBC, hDBC, ndx, @sqlState, nativeError, @msg, SQL_MAX_MESSAGE_LENGTH, msgLength);
    if((rc = SQL_SUCCESS) or (rc = SQL_SUCCESS_WITH_INFO))then
    begin
      SetString(eState, sqlState, StrLen(sqlState));
      if(eState = 'HY000')then
        rc := SQLGetDiagField(SQL_HANDLE_DBC, hDBC, ndx, SQL_DIAG_MESSAGE_TEXT, @msg, SQL_MAX_MESSAGE_LENGTH, msglength);
    end;

    // Test to see if there are any "Diag" records to show.
    Result := ((rc = SQL_SUCCESS) or (rc = SQL_SUCCESS_WITH_INFO));
    if(Result)then
    begin
      // Ignore if error is:  "Driver doesn't support...".
      if(eState = 'IM001')then
        Result := False
      else
      begin
        SetString(errMsg, msg, StrLen(msg));
        errMsg := errMsg + '. State:  ' + eState;
      end;
    end;
    // sets a flag when certain messages or error states are found
    // that may indicate a lost connection
    CheckConnFail(eState, errMsg);
  end
  else
    Result := False;
end;

end.
