unit sql;

interface

uses
  Classes, SysUtils;

type
  // Possible kinds of data sources, like Oracle, XBase and Access respectfully.
  TdbType = (dbTypeUnknown, dbTypeOracle, dbTypeMySQL, dbTypeDBF, dbTypeFireBird, dbAccess, dbExcel, dbSqlServer, dbCodeBase);

  TSql = class       // a class for creating INSERT and UPDATE SQL strings...
  private
    { Private declarations }

    // added support for different database types, default is Oracle
    fdbType: TdbType;                     // present database type
    fInWhere: boolean;                    // building a where clause?
    fDelimWithBrackets : boolean;         // delimit all fields with []
    ColumnList : TStringList;             // list of column names
    ValueList : TStringList;              // list of values (associated to above)
    CompareList : TStringList;            // list of comparison operators (for buildwhere only)
    AndOrList: TStringList;               // list of and/or 's for above

    QuotedList : array of boolean;        // quote flags for above pairs

    function FixUp(Value: string): string;      // fix up the quotes etc...
    function FieldFixUp(Field: string): string;
    function GetCount: integer;
    function GetValue(ndx: integer): string;

  public
    { Public declarations }
    property InWhere: boolean read fInWhere write fInWhere;
    property DelimWithBrackets : boolean read fDelimWithBrackets write fDelimWithBrackets;
    property Count: integer read GetCount;
    property Value[ndx: integer]: string read GetValue;

    function FindIndex(colname: string): integer;
    // any query added yet?
    function AnyQuery: boolean;

    // add column value pair
    function Add(colname, value: string; quotes: boolean = true): integer;	overload;
    function Add(colname: string; i: integer): integer; overload;
    function Add(colname: string; f: double): integer; overload;
    function AddC(colName: string; c: currency): integer; overload;
    function Add(colname: string; b: boolean): integer; overload;
    // add a date field with the to_date() func...
    function Add(colname, value, dateFormat: string): integer; overload;
    function AddTd(colname: string; datetime: TDateTime): integer;
    function AddDt(colName : string; datetime: TDateTime): integer;
    function AddTm(colname: string; datetime: TDateTime): integer;
    function AddYN(colname: string; b: boolean): integer;
    function AddNow(colName: string): integer;

    procedure Inc(colname, value: string; FormatString: string = '%12.2f');

    // change a value that was already set...
    procedure Change(colname, value: string; quotes: boolean = true); overload;
    procedure Change(colname: string; i: integer; quotes: boolean = false); overload;
    procedure Change(colname: string; d: double; quotes: boolean = false); overload;

    procedure ChangeDt(colName: string; datetime: TDateTime);
    procedure ChangeQ(colname: string; quotes: boolean = False);

    procedure Copy(sql: TSql);

    // add more data to existing field
    procedure Addto(colname, value: string; quotes: boolean = true);

    // delete a column from the lists...
    procedure Remove(colName: string);
    function FindAndRemove(colName: string): string;

    procedure SetCompare(colndx: integer; Comparison: string; AndOr: string = 'and'); overload;
    procedure SetCompare(colName, Comparison: string; AndOr: string = 'and'); overload;

    // find a value in amongst the pairs
    function FindValue(colname: string): string;

    // function to return a string that, when executed, will return the last auto incremented field value...
    function LastAutoInc(field: string; seqname: string = ''): string;

    // clear the list of pairs
    procedure Clear(InWhere: boolean = False);

    // build either insert or update based on insert boolean...
    function BuildCmd(table: string; insert: boolean = True; suffix: string = ''): string;
    // sql Update statement
  	function BuildUpdate(prefix: string; suffix: string = ''): string;
    // sql Insert statements
    function BuildInsert(prefix: string; mid: string = 'Values'; suffix: string = ''): string;
    // Where part of an sql statement...
    function BuildWhere(prefix: string = ''; suffix: string = ''):string;

    constructor Create(dbType: TdbType = dbTypeOracle);    // default is oracle
    destructor Destroy; override;
  end;


implementation

// constructor for Class that will handle SQL query Update/insert strings
constructor TSql.Create(dbType: TdbType = dbTypeOracle);    // default is oracle
begin
  inherited Create;    { must call the ancestors create }

  fdbType := dbType;
  fInWhere := False;
  fDelimWithBrackets := False;

  ColumnList := TStringList.Create;
  ValueList := TStringList.Create;
  CompareList := TStringList.Create;
  AndOrList := TStringList.Create;

  SetLength(QuotedList, 0);   // init to nothing
end;

// cleanup, destroy all stringlists
destructor TSql.Destroy;
begin
  SetLength(QuotedList, 0);
  QuotedList := nil;

  AndOrList.Free;
  ColumnList.Free;
  ValueList.Free;
  CompareList.Free;

  inherited Destroy;    { call the Ancestors destroy }
end;

// return if anything has been added to the query yet
function TSql.AnyQuery: boolean;
begin
  Result := ColumnList.Count > 0;
end;

function TSql.GetCount: integer;
begin
  Result := ColumnList.Count;
end;

function TSql.GetValue(ndx: integer): string;
begin
  if((ndx >= 0) and (ndx < ValueList.Count))then
    Result := ValueList.Strings[ndx]
  else
    Result := '';
end;

// clear out the list of columnname/value pairs
procedure TSql.Clear(InWhere: boolean);
begin
  ColumnList.Clear;
  ValueList.Clear;
  CompareList.Clear;
  SetLength(QuotedList, 0);
  QuotedList := nil;

  fInWhere := InWhere;
end;

// find the columns index
function TSql.FindIndex(colName: string): integer;
begin
  Result := ColumnList.IndexOf(LowerCase(colName));  // force lower case search
end;

// Find Value for the given column name
function TSql.FindValue(colName: string): string;
var
  ndx: integer;
begin
  ndx := FindIndex(colName);    // look for column name
  if(ndx <> -1)then
    Result := ValueList.Strings[ndx]
  else
    Result := '';   // couldn't find it
end;

// add a fieldname and value pair
// to the querys VarList
function TSql.Add(colName, value: string; quotes: boolean): integer;
var
  v: string;
begin
  Result := FindIndex(colName);    // make sure doesn't already exist
  if((Result = -1) or (fInWhere))then            // must allow duplicates when building Where
  begin
    v := Trim(value);
    if((Length(v) = 0) and not quotes)then
      v := 'null';

    Result := ColumnList.Add(LowerCase(colName));     // add the column name

    ValueList.Add(v);                                 // add the value, check quotes
    SetLength(QuotedList, ColumnList.Count);          // add one to the quote flags array
    QuotedList[ColumnList.Count - 1] := quotes;       // or (Length(v) = 0);  // set the flag
  end;
end;

//  same as above but add an integer value
function TSql.Add(colName: string; i : integer): integer;
begin
  Result := Add(colName, IntToStr(i), false);
end;

//  same as above but add an extened value
function TSql.Add(colName: string; f: double): integer;
begin
  Result := Add(colName, FloatToStr(f), false);
end;

//  same as above but add an currency value
function TSql.AddC(colName: string; c: currency): integer;
begin
  Result := Add(colName, CurrToStr(c), false);
end;

//  same as above but add boolean support
function TSql.Add(colName: string; b: boolean): integer;
begin
  if b then
    Result := Add(colName, 1)
  else
    Result := Add(colName, 0);
end;

// same as above but add a date string using the to_date() function
function TSql.Add(colName, value, dateFormat: string): integer;
begin
  if(fdbType = dbTypeOracle)then
    value := 'to_date(''' + value + ''', ''' + dateFormat + ''')'
  else
    if(fdbType = dbTypeDBF)then
      value := '{' + value + '}';

  Result := Add(colName, value, false);
end;

// same as above but adds a boolean as Y or N
function TSql.AddYN(colname: string; b: boolean): integer;
const
  BoolStr: array[boolean] of string = ('N', 'Y');
begin
  Result := Add(colname, BoolStr[b]);
end;

// add date only!!
function TSql.AddDt(colName : string; datetime : TDateTime): integer;
begin
(*
  case fdbType of
    dbTypeOracle:
      Result := Add(colName, FormatDateTime('yyyymmdd', datetime), 'YYYYMMDD');

    dbTypeDBF:
      Result := Add(colName, '{' + FormatDateTime('MM-DD-YYYY', datetime) + '}', False);

    dbTypeUnknown:
        Result := -1

    else
      // this method works for all (so far?!)
      Result := Add(colName, '{d ''' + FormatDateTime('YYYY-MM-DD', datetime) + '''}', False);
  end;
*)
  // this method works for all (so far?!)
  Result := Add(colName, '{d ''' + FormatDateTime('YYYY-MM-DD', datetime) + '''}', False);
end;

// same as above but add a time date stamp
function TSql.AddTd(colName : string; datetime : TDateTime): integer;
begin
(*
  case fdbType of
    dbTypeOracle:
      Result := Add(colName, FormatDateTime('yyyymmdd hhnnss', datetime), 'YYYYMMDD HH24MISS');

    dbTypeDBF:
      Result := Add(colName, '{' + FormatDateTime('MM-DD-YYYY HH:NN:SS', datetime) + '}', False);

    dbTypeUnknown:
      Result := -1

    else
      Result := Add(colName, '{ts ''' + FormatDateTime('YYYY-MM-DD HH:NN:SS', datetime) + '''}', False);
  end;
*)
  Result := Add(colName, '{ts ''' + FormatDateTime('YYYY-MM-DD HH:NN:SS', datetime) + '''}', False);
end;

function TSql.AddTm(colName: string; datetime : TDateTime): integer;
begin
  Result := Add(colName, FormatDateTime('HH:NN:SS', datetime));
// this would be for a time type field, but I think we just want to
// put the time in a character field, right?
//  Result := Add(colName, '{t ''' + FormatDateTime('HH:NN:SS', datetime) + '''}', False);
end;

// add present time/date, system independent
function TSql.AddNow(colName : string): integer;
begin
(*
  case fdbType of
    dbTypeMySQL:
      Result := Add(colName, 'CURRENT_TIMESTAMP()', false);
    dbTypeOracle:
      Result := Add(colName, 'SYSDATE', false);
  else
    Result := Add(colName, '{fn now()}', False);
  end;
*)
  case fdbType of
    dbCodeBase:       // codebase doesn't appear to have a datetime datatype support...
      Result := Add(colName, '{fn curdate()}', False);  // curdate() is date part only, no time...
  else
    Result := Add(colName, '{fn now()}', False);
  end;
end;

// add to an existing value string
procedure TSql.AddTo(colName, value: string; quotes: boolean);
var
  ndx: integer;
begin
  ndx := FindIndex(colName);
  if(ndx <> -1)then
  begin
    ValueList.Strings[ndx] := ValueList.Strings[ndx] + value;  // append
    QuotedList[ndx] := quotes; // or (Length(ValueList.Strings[ndx]) = 0);  // set the flag
  end
  else
    Add(colName, value, quotes);
end;

// add value to what is already there ('value' must be string number)
procedure TSql.Inc(colName, value, FormatString: string);
var
  ndx: integer;
begin
  ndx := FindIndex(colName);
  if(ndx <> -1)then
    ValueList.Strings[ndx] := Trim(Format(FormatString, [StrToFloat(value) + StrToFloat(ValueList.Strings[ndx])]))
  else
    Add(colName, value);
end;

//  change an existing value, or add it if it's not there...
procedure TSql.Change(colName, value: string; quotes: boolean);
var
  v: string;
  ndx: integer;
begin
  ndx := FindIndex(colName);
  if(ndx = -1)then
    Add(colName, value, quotes)
  else
  begin
    v := Trim(value);
    if((Length(v) = 0) and not quotes)then
      v := 'null';

    ValueList.Strings[ndx] := v;
    QuotedList[ndx] := quotes;
  end;
end;

procedure TSql.ChangeDt(colName: string; datetime: TDateTime);
var
  ndx: integer;
begin
  ndx := FindIndex(colName);
  if(ndx = -1)then
    AddDt(colName, datetime)
  else
    ValueList.Strings[ndx] := '{d ''' + FormatDateTime('YYYY-MM-DD', datetime) + '''}';
end;

// same as above, but value is passed as a integer number...
procedure TSql.Change(colname: string; i: integer; quotes: boolean);
begin
  Change(colname, IntToStr(i), quotes);
end;

// same as above, but value is passed as a double number...
procedure TSql.Change(colname: string; d: double; quotes: boolean);
begin
  Change(colname, FloatToStr(d), quotes);
end;

// change the quotes flag for a field...
procedure TSql.ChangeQ(colname: string; quotes: boolean = False);
var
  ndx: integer;
begin
  ndx := FindIndex(colName);
  if(ndx <> -1)then
    QuotedList[ndx] := quotes;
end;

// copy all fields and data from another TSql...
procedure TSql.Copy(sql: TSql);
var
  i: integer;
begin
  for i := 0 to sql.ColumnList.Count - 1 do
    Add(sql.ColumnList.Strings[i], sql.ValueList.Strings[i], sql.QuotedList[i]);
end;

// take one out of the lists...
procedure TSql.Remove(colName: string);
var
  ndx, i: integer;
begin
  ndx := FindIndex(colName);
  if(ndx <> -1)then
  begin
    ColumnList.Delete(ndx);                     // remove the column name
    ValueList.Delete(ndx);                      // and the value...
    for i := ndx to ColumnList.Count - 1 do
      QuotedList[i] := QuotedList[i + 1];       // shift the flags down one
    SetLength(QuotedList, ColumnList.Count);    // subtract one from the quote flags array
  end;
end;

// Find Value and Remove for the given column name
function TSql.FindAndRemove(colName: string): string;
begin
  Result := FindValue(colName);
  Remove(colName);
end;

procedure TSql.SetCompare(colndx: integer; Comparison: string; AndOr: string);
begin
  if(colndx < ColumnList.Count)then
  begin
    // we are setting a compare value, must in InWhere!?
    fInWhere := True;

    // make sure we have enough holding places to change the required spot...
    while(CompareList.Count <= colndx)do
      CompareList.Add('=');               // default is; compare for equal...
    CompareList.Strings[colndx] := Comparison;

    // if we are only on the first (colndx=0) comparison, what do we and / or it with?
    if(colndx > 0)then
    begin
      // make sure we have enough holding places to change the required spot...
      while(AndOrList.Count <= colndx)do
        AndOrList.Add('and');               // default is; compare for equal...
      AndOrList.Strings[colndx] := AndOr;
    end;
  end;
end;

// set the comparison value for a field (used in BuildWhere)
procedure TSql.SetCompare(colName, Comparison: string; AndOr: string);
var
  ndx: integer;
begin
  ndx := FindIndex(colName);
  if(ndx <> -1)then
    SetCompare(ndx, Comparison, AndOr);
end;

function TSql.LastAutoInc(field: string; seqname: string): string;
begin
  case fdbType of
    dbTypeOracle:
      // for Oracle, we need to know the sequence name...
      Result := 'select ' + seqname + '.currval as ' + field + ' from dual';
  else
    // this has been tested on several data sources, (Access, SQL Server, MySQL) but not all...
    Result := 'SELECT @@IDENTITY AS ' + field
  end;
end;

function TSql.FixUp(Value: string): string;
begin
  // Double up any quotes in the string...
  Result := StringReplace(Value, '''', '''''', [rfReplaceAll]);

  // rb Aug 24/05, looks like MySQL demands \\ for \
  if(fdbType = dbTypeMySQL)then
    Result := StringReplace(Result, '\', '\\', [rfReplaceAll]);
end;

function TSql.FieldFixUp(Field: string): string;
begin
  Result := Field;

  // new rb Dec 12/2011, added optional delim with []
  if fDelimWithBrackets then
  begin
    if(System.Copy(Result, 1, 1) <> '[')then
      Result := '[' + Result + ']';
  end
  else
  begin
    // if there is a space or a - in field name, make sure double quotes surround it...
    if((Pos(' ', Result) > 0) or (Pos('-', Result) > 0))then
      if(System.Copy(Result, 1, 1) <> '"')then
        Result := '"' + Result + '"';
  end;
  // can't seem to find a way to deal with '?' in field name!?!
  //Result := StringReplace(Result, '?', '{?}', [rfReplaceAll]);
end;

function TSql.BuildCmd(table: string; insert: boolean = True; suffix: string = ''): string;
begin
  if(insert)then
    Result := BuildInsert('insert into ' + table)
  else
    Result := BuildUpdate('update ' + table, suffix);
end;

// build an SQL update statement
// from the list of fields and data in the StringList pairs
function TSql.BuildUpdate(prefix: string; suffix: string = ''):string;
var
  sqlcmd, v: string;
  i: integer;
begin
  sqlcmd := prefix + ' set ';   // start with prefix

  for i := 0 to ColumnList.Count - 1 do
  begin
		if i <> 0 then
      sqlcmd := sqlcmd + ', ';    // need comma if not first one

    sqlcmd := sqlcmd + FieldFixUp(ColumnList.Strings[i]) + '=';
    v := ValueList.Strings[i];
    if(QuotedList[i] or (Length(v) = 0))then
    begin
      sqlCmd := sqlCmd + '''';
      sqlcmd := sqlcmd + FixUp(v);
      sqlCmd := sqlCmd + '''';
    end
    else
      sqlcmd := sqlcmd + v;
  end;

  if(Length(suffix) > 0)then
    sqlcmd := sqlcmd + ' ' + suffix;
  Result := sqlcmd;
end;

// build an SQL command from the linked list of
// fields and variables
function TSql.BuildInsert(prefix, mid, suffix: string):string;
var
  sqlcmd, v: string;
  i: integer;
begin
  sqlcmd := prefix + ' (';   // start with prefix

  for i := 0 to ColumnList.Count - 1 do
  begin
		if i <> 0 then
      sqlcmd := sqlcmd + ', ';    // need comma if not first one
    sqlcmd := sqlcmd + FieldFixUp(ColumnList.Strings[i]);
  end;

	sqlcmd := sqlcmd + ') ' + mid + ' (';
  for i := 0 to ColumnList.Count - 1 do
  begin
		if i <> 0 then
      sqlcmd := sqlcmd + ', ';    // need comma if not first one
    v := ValueList.Strings[i];
    if(QuotedList[i] or (Length(v) = 0))then
    begin
      sqlCmd := sqlCmd + '''';
      sqlcmd := sqlcmd + FixUp(v);
      sqlCmd := sqlCmd + '''';
    end
    else
      sqlcmd := sqlcmd + v;
  end;

	Result := sqlcmd + ')' + suffix;
end;

// and a where clause?
// changed, you can now set the AndOrList to something other than AND
// === fixed May/07  ===: only good if you need to AND *ALL* conditions...
function TSql.BuildWhere(prefix, suffix: string):string;
var
  sqlcmd, v, Compare, Conjunct: string;
  i: integer;
begin
  sqlcmd := Trim(prefix);
  if(Length(sqlcmd) > 0)then
    sqlcmd := sqlcmd + ' ';

  for i := 0 to ColumnList.Count - 1 do
  begin
		if(i <> 0)then
    begin
      // they may not have set the conjuction operator, default to 'and' ...
      if(i < AndOrList.Count)then
        Conjunct := AndOrList.Strings[i]
      else
        Conjunct := 'and';

      sqlcmd := sqlcmd + ' ' + Conjunct + ' ';
    end;

    // they may not have set the comparison operator, default to '=' ...
    if(i < CompareList.Count)then
      Compare := CompareList.Strings[i]
    else
      Compare := '=';

    v := ValueList.Strings[i];

    if((Compare = '=') and (LowerCase(v) = 'null'))then
      Compare := ' is ';

    sqlcmd := sqlcmd + FieldFixUp(ColumnList.Strings[i]) + Compare;
    if(QuotedList[i] or (Length(v) = 0))then
    begin
      sqlCmd := sqlCmd + '''';
      sqlcmd := sqlcmd + FixUp(v);
      sqlCmd := sqlCmd + '''';
    end
    else
      sqlcmd := sqlcmd + v;
  end;

  // may never need this, but it's here...
  if(Length(suffix) > 0)then
    sqlcmd := sqlcmd + ' ' + suffix;
  Result := sqlcmd;
end;

end.
